from crikit.fe import *
from crikit.fe_adjoint import *
from crikit.observer import SurfaceObserver
from crikit.observe_block import observe_error
from pyadjoint_utils import Block, function_get_local
from pyadjoint_utils.jax_adjoint import array, ndarray, overload_jax
from pyadjoint.tape import annotate_tape, stop_annotating, get_working_tape
from crikit.cr.jax_utils import (
    JAXArrays,
    UFLExprSpace_To_JAX,
    JAX_UFLFunctionSpace_Covering,
)
from crikit.cr.ufl import UFLFunctionSpace
from crikit.cr.types import PointMap
from crikit.cr.space_builders import DirectSum
from crikit.covering import get_composite_cr
import jax
from jax import numpy as jnp
from jax.tree_util import Partial as partial
from jax import random
import numpy as np
from typing import Union, Optional


def integral_loss(
    meas: Union[Function, SurfaceObserver.SurfaceObservation], pred: Function
) -> AdjFloat:
    """Computes the squared L^2 loss between ``meas`` and ``pred``.

    :param meas: The observed data
    :type meas: Union[Function, SurfaceObserver.SurfaceObservation]
    :param pred: The observed data
    :type pred: Function
    :return: The integral of the squared difference between ``meas`` and ``pred``
    :rtype: AdjFloat
    """
    if isinstance(meas, SurfaceObserver.SurfaceObservation):
        # Integrate with the specified infinitesimal element.
        diff = pred.u - meas.u
        return assemble(inner(diff, diff) * meas.de)
    elif isinstance(meas, Function):
        # Integrate over the entire domain.
        diff = pred - meas
        return assemble(inner(diff, diff) * dx)


def vector_loss(meas, pred):
    return observe_error(pred, meas.vector())


class SlicedWassersteinDistance(PointMap):
    """A class that computes the sliced p-th Wasserstein distance, which is
    obtained by projecting the data into ``num_projections`` random directions,
    then computing the 1-dimensional p-th Wasserstein distances of the sliced
    data and averaging those.
    """

    def __init__(
        self,
        V: FunctionSpace,
        num_projections: int,
        key: int,
        p: int = 2,
        u_weights: Optional[jnp.ndarray] = None,
        v_weights: Optional[jnp.ndarray] = None,
    ):
        """

        :param V: The :class:`crikit.fe_adjoint.FunctionSpace` in which the
        inputs to this class live.
        :type V: crikit.fe_adjoint.FunctionSpace
        :param num_projections: How many random directions to project the data in?
        :type num_projections: int
        :param key: the key for :mod:`jax.random` to use for sampling (i.e. result of a call to `jax.random.PRNGKey(seed)`)
        :type key: int
        :param p: the "p" in "p-th Wasserstein distance", defaults to 2
        :type p: int, optional
        :param u_weights: Optional weights to use for the first distribution passed to this class, defaults to None
        :type u_weights: jax.numpy.ndarray, optional
        :param v_weights: Optional weights to use for the second distribution passed to this class, defaults to None
        :type v_weights: jax.numpy.ndarray, optional
        """

        ufl_u = Function(V)
        qp_shape = ufl_u.ufl_shape
        n_qp = ufl_u.vector().vec().array.size // np.prod(qp_shape)
        jax_u = jnp.ones((n_qp, *qp_shape))
        self._backend = BackendSlicedWassersteinDistance(
            jax_u,
            jax_u,
            num_projections,
            key,
            u_weights=u_weights,
            v_weights=v_weights,
            p=p,
        )

        self._backend = get_composite_cr(
            DirectSum(UFLFunctionSpace(V), UFLFunctionSpace(V)), self._backend
        )

        super().__init__(self._backend.source, self._backend.target)

    def __call__(self, u: Function, v: Function) -> ndarray:
        """Computes the p-th sliced Wasserstein distance between ``u`` and ``v``.
        :param u: The values of the first distribution
        :type u: crikit.fe_adjoint.Function
        :param v: The values of the second distribution
        :type v: crikit.fe_adjoint.Function
        :return: The p-th sliced Wasserstein distance between ``u`` and ``v``
        :rtype: pyadjoint_utils.jax_adjoint.ndarray
        """
        return self._backend((u, v))


class BackendSlicedWassersteinDistance(PointMap):
    """This class computes the sliced Wasserstein distance with JAX
    inputs and outputs. It's intended to be put inside a driver that calls
    get_composite_cr() on it so it can take UFL inputs.
    """

    def __init__(self, u, v, num_projections, key, u_weights=None, v_weights=None, p=2):
        self._u = u
        self._v = v
        self._key = key
        self._n = int(num_projections)
        self._samples = [jnp.eye(1)] if u.shape[1] == 1 else []
        if u.shape[1] > 1:
            for i in range(self._n):
                self._key, subkey = random.split(self._key)
                d = random.normal(subkey, shape=(u.shape[1],))
                self._samples.append(d / jnp.linalg.norm(d))

        self._u_w = u_weights
        self._v_w = v_weights
        self._args = ()
        if self._u_w is not None:
            if self._v_w is None:
                self._v_w = jnp.ones_like(v)
            self._distance = (
                weighted_sliced_wasserstein_distance
                if p == 1
                else lambda u, v, d, uw, vw: weighted_sliced_wasserstein_p_distance(
                    u, v, d, uw, vw, p
                )
            )
            self._args = (self._u_w, self._v_w)
        elif self._v_w is not None:
            self._u_w = jnp.ones_like(u)
            self._distance = (
                weighted_sliced_wasserstein_distance
                if p == 1
                else lambda u, v, d, uw, vw: weighted_sliced_wasserstein_p_distance(
                    u, v, d, uw, vw, p
                )
            )
            self._args = (self._u_w, self._v_w)
        else:
            self._distance = (
                sliced_wasserstein_distance
                if p == 1
                else lambda u, v, d: sliced_wasserstein_p_distance(u, v, d, p)
            )

        in_types = DirectSum(JAXArrays(u.shape), JAXArrays(v.shape))
        out_type = JAXArrays((jnp.mean(jnp.array([0, 1]))).shape)
        super().__init__(in_types, out_type)

    @property
    def num_projections(self):
        return self._n

    @num_projections.setter
    def num_projections(self, n):
        self._n = n

    @property
    def u_weights(self):
        return self._u_w

    @u_weights.setter
    def u_weights(self, new_weights):
        self._u_w = new_weights
        if self._v_w is None:
            self._v_w = jnp.ones_like(self._v)
            self._distance = weighted_sliced_wasserstein_distance
            self._args = (self._u_w, self._v_w)

    @property
    def v_weights(self):
        return self._v_w

    @u_weights.setter
    def v_weights(self, new_weights):
        self._v_w = new_weights
        if self._u_w is None:
            self._u_w = jnp.ones_like(self._u)
            self._distance = weighted_sliced_wasserstein_distance
            self._args = (self._u_w, self._v_w)

    def __call__(self, val):
        u, v = val
        return self._distance(u, v, jnp.array(self._samples), *self._args)


@partial(overload_jax, argnums=(0, 1))
def sliced_wasserstein_distance(u, v, directions, *args):
    def body_fun(val, d):
        (u, v) = val
        uproj = u @ d
        vproj = v @ d
        return (u, v), wasserstein_distance_1d(uproj, vproj)

    (u, v), distances = jax.lax.scan(body_fun, (u, v), directions)
    return jnp.mean(distances)


@partial(overload_jax, argnums=(0, 1))
def weighted_sliced_wasserstein_p_distance(
    u, v, directions, uweights, vweights, p, *args
):
    def body_fun(val, d):
        (u, v, uw, vw) = val
        uproj = u @ d
        vproj = v @ d

        return (u, v, uw, vw), _weighted_cdf_distance(p, uproj, vproj, uw, vw)

    (u, v), distances = jax.lax.scan(body_fun, (u, v, uw, vw), directions)
    return jnp.mean(distances)


@partial(overload_jax, argnums=(0, 1))
def sliced_wasserstein_p_distance(u, v, directions, p, *args):
    def body_fun(val, d):
        (u, v, p) = val
        uproj = u @ d
        vproj = v @ d
        return (u, v, p), _cdf_distance(uproj, vproj, p)

    (u, v, p), distances = jax.lax.scan(body_fun, (u, v, p), directions)
    return jnp.mean(distances)


@partial(overload_jax, argnums=(0, 1))
def weighted_sliced_wasserstein_distance(u, v, directions, uweights, vweights, *args):
    def body_fun(val, d):
        (u, v, uw, vw) = val
        uproj = u @ d
        vproj = v @ d

        return (u, v, uw, vw), weighted_wasserstein_distance_1d(uproj, vproj, uw, vw)

    (u, v), distances = jax.lax.scan(body_fun, (u, v, uw, vw), directions)
    return jnp.mean(distances)


# implementation of the Wasserstein distance taken from scipy
def wasserstein_distance_1d(u, v):
    """1-D Wasserstein distance between two distributions
    :param u: The first distribution
    :type u: jnp.ndarray
    :param v: The second distribution
    :type v: jnp.ndarray
    :return: The Wasserstein distance between ``u`` and ``v``.
    :rtype: jnp.ndarray
    """
    u_sort = jnp.argsort(u)
    v_sort = jnp.argsort(v)

    values = jnp.sort(jnp.concatenate((u, v)))

    deltas = jnp.diff(values)
    u_idx = u[u_sort].searchsorted(values[:-1], "right")
    v_idx = v[v_sort].searchsorted(values[:-1], "right")

    u_cdf = u_idx / u.size
    v_cdf = v_idx / v.size
    return jnp.sum(jnp.multiply(jnp.abs(u_cdf - v_cdf), deltas))


def weighted_wasserstein_distance_1d(u, v, uweights, vweights):
    """1-D Wasserstein distance between two distributions
    :param u: The first distribution
    :param v: The second distribution
    :param uweights: The weights
    :return: The Wasserstein distance between ``u`` and ``v``.
    :rtype: jnp.ndarray
    """
    u_sort = jnp.argsort(u)
    v_sort = jnp.argsort(v)

    values = jnp.sort(jnp.concatenate((u, v)))

    deltas = jnp.diff(values)

    u_idx = u[u_sort].searchsorted(values[:-1], "right")
    v_idx = v[v_sort].searchsorted(values[:-1], "right")

    u_sort_cum = jnp.concatenate(([0], jnp.cumsum(uweights[u_idx])))
    v_sort_cum = jnp.concatenate(([0], jnp.cumsum(vweights[v_idx])))
    u_cdf = u_sort_cum[u_idx] / u_sort_cum[-1]
    v_cdf = v_sort_cum[v_idx] / v_sort_cum[-1]
    return jnp.sum(jnp.multiply(jnp.abs(u_cdf - v_cdf), deltas))


@jax.jit
def _cdf_distance(u, v, p):
    u_sort = jnp.argsort(u)
    v_sort = jnp.argsort(v)

    values = jnp.sort(jnp.concatenate((u, v)))

    deltas = jnp.diff(values)
    u_idx = u[u_sort].searchsorted(values[:-1], "right")
    v_idx = v[v_sort].searchsorted(values[:-1], "right")

    u_cdf = u_idx / u.size
    v_cdf = v_idx / v.size
    return jnp.power(
        jnp.sum(jnp.multiply(jnp.power(jnp.abs(u_cdf - v_cdf), p), deltas)), 1 / p
    )


@jax.jit
def _weighted_cdf_distance(p, u, v, uweights, vweights):
    u_sort = jnp.argsort(u)
    v_sort = jnp.argsort(v)

    values = jnp.sort(jnp.concatenate((u, v)))

    deltas = jnp.diff(values)

    u_idx = u[u_sort].searchsorted(values[:-1], "right")
    v_idx = v[v_sort].searchsorted(values[:-1], "right")

    u_sort_cum = jnp.concatenate(([0], jnp.cumsum(uweights[u_idx])))
    v_sort_cum = jnp.concatenate(([0], jnp.cumsum(vweights[v_idx])))
    u_cdf = u_sort_cum[u_idx] / u_sort_cum[-1]
    v_cdf = v_sort_cum[v_idx] / v_sort_cum[-1]

    return jnp.power(jnp.sum(jnp.multiply(jnp.power(jnp.abs(u - v), p), deltas)), 1 / p)



from pyadjoint import minimize as pyadjoint_minimize
from .callback import Callback, FileLoggerCallback, CallbackCombiner
from typing import Sequence, Callable, Union, Optional


class _default:
    pass


def minimize(
    rf,
    method="L-BFGS-B",
    scale=1.0,
    callbacks=_default(),
    **kwargs,
):
    """CRIKit's wrapper around `pyadjoint`'s `minimize` function that
    itself calls `scipy`'s minimization routines. See pyadjoint's
    documentation for information on `kwargs`, which are passed through
    to pyadjoint's `minimize` function.

    :param rf: The functional to optimize
    :type rf: ReducedFunctional
    :param method: The optimization method used, defaults to `'L-BFGS-B'`
    :type method: str
    :param scale: Scale the loss functional by this amount, defaults to `1.0`
    :type scale: float
    :param callbacks: A list or tuple of callback functions or classes
    to combine into one callback, or a single such callback function,
    or None if no callback is desired. Defaults to a
    {class}`FileLoggerCallback` with default arguments
    :type callbacks: Optional[Sequence[Union[Callable, Callback]]]

    """
    if isinstance(callbacks, _default):
        if "callback" in kwargs:
            callbacks = kwargs["callback"]
        else:
            callbacks = FileLoggerCallback(rf=rf)
    if isinstance(callbacks, (list, tuple)):
        callback = CallbackCombiner(callbacks)
    elif callable(callbacks):
        if isinstance(callbacks, FileLoggerCallback):
            # in the default case, make sure the callback has acccess to
            # the rf being minimized. If the user has passed
            # `write_params=False` to its ctor, this will have no effect
            if callbacks.rf is None:
                callbacks.rf = rf
        callback = callbacks
    else:
        callback = callbacks
    kwargs["callback"] = callback

    return pyadjoint_minimize(rf, method=method, scale=scale, **kwargs)



from autograd import vector_jacobian_product, jacobian
from autograd.numpy.numpy_boxes import ArrayBox
from autograd.numpy.numpy_vspaces import ArrayVSpace
from pyadjoint.enlisting import Enlist
from pyadjoint.overloaded_type import create_overloaded_object
from pyadjoint.tape import annotate_tape, get_working_tape, stop_annotating
from pyadjoint_utils import Block, JacobianIdentity
from pyadjoint import AdjFloat
from pyadjoint_utils.numpy_adjoint import ndarray
from itertools import product
from functools import wraps
import numpy as np

ArrayBox.register(ndarray)
ArrayBox.register(AdjFloat)
ArrayVSpace.register(ndarray)
ArrayVSpace.register(AdjFloat)


def overload_autograd(func, pointwise):
    """Create an overloaded version of an Autograd function.

    This method makes several assumptions:
    1) The function is explicit, i.e. ``y = func(x)``, where y is the output of the operation.
    2) All of y is possible to convert to an OverloadedType.
    3) Unless annotation is turned off, the operation should always be annotated when calling the overloaded function.

    After the overloaded function is called, the pointwise bool for each input
    is recorded in the function's pointwise attribute.

    Args:
        func (function): The target function for which to create an overloaded version.
        pointwise (bool or list[bool]): True means the function performs
            operations on a batch of points. This allows optimizing the Jacobian
            calculations by only computing the diagonal component. If a list,
            then there should be a bool for each input.
    Returns:
        function: An overloaded version of ``func``
    """

    @wraps(func)
    def _overloaded_function(*args, **kwargs):
        annotate = annotate_tape(kwargs)
        with stop_annotating():
            func_output = func(*args, **kwargs)

        func_output = Enlist(func_output)

        if len(func_output) > 1:
            raise ValueError(
                "overload_autograd currently only works with functions that output a single numpy array"
            )

        r = [create_overloaded_object(out) for out in func_output]
        if annotate:
            block = AutogradBlock(func, args, r, kwargs, pointwise=pointwise)
            _overloaded_function.pointwise = block._pointwise
            tape = get_working_tape()
            tape.add_block(block)

        return func_output.delist(r)

    _overloaded_function.pointwise = pointwise
    return _overloaded_function


def overloaded_autograd(pointwise):
    """Returns a decorator for Autograd functions that should be overloaded

    Args:
        pointwise (bool or list[bool]): See :func:`overload_autograd`.
    """

    def decorator(func):
        return overload_autograd(func, pointwise)

    return decorator


class AutogradBlock(Block):
    def __init__(self, func, args, outputs, func_kwargs, pointwise):
        super().__init__()

        self._func = func
        self._kwargs = func_kwargs

        nterms = len(args)

        if isinstance(pointwise, bool):
            self._pointwise = (pointwise,) * nterms
        else:
            self._pointwise = pointwise
        if len(self._pointwise) != nterms:
            raise ValueError(
                f"len(pointwise) != nterms. For each input, you must specify if it is defined pointwise. ({len(self._pointwise)} != {nterms})"
            )

        self._arg_shapes = [
            a.shape[1:] if pw else a.shape for a, pw in zip(args, self._pointwise)
        ]
        self._out_pointwise = any(self._pointwise)
        self._out_shape = (
            outputs[0].shape[1:] if self._out_pointwise else outputs[0].shape
        )

        self._vjpfuncs = [
            vector_jacobian_product(self._func, argnum=i) for i in range(nterms)
        ]

        # there is an autograd-forward project, but afaik it is not integrated yet, so we use the
        # twice-backwards trick with a dummy argument (Jamie Townsend,
        # https://j-towns.github.io/2017/06/12/A-new-trick.html)
        self._vjp2funcs = [
            vector_jacobian_product(self._vjpfuncs[i], argnum=nterms)
            for i in range(nterms)
        ]
        self._hesfuncs = [
            vector_jacobian_product(self._vjpfuncs[i], argnum=j)
            for i, j in product(range(nterms), range(nterms))
        ]

        self._tlm_matrix_funcs = [jacobian(self._func, argnum=i) for i in range(nterms)]

        for arg in args:
            self.add_dependency(arg)

        for out in outputs:
            self.add_output(out.block_variable)

    def __repr__(self):
        return f"AutogradBlock({self._func})"

    def prepare_recompute_component(self, inputs, relevant_outputs):
        return Enlist(self._func(*inputs, **self._kwargs))

    def recompute_component(self, inputs, block_variable, idx, prepared):
        return prepared[idx]

    def evaluate_adj_component(
        self, inputs, adj_inputs, block_variable, idx, prepared=None
    ):
        if len(adj_inputs) == 1:
            adj_inputs = adj_inputs[0]
        return self._vjpfuncs[idx](*inputs, adj_inputs, **self._kwargs)

    def prepare_evaluate_tlm(self, inputs, tlm_inputs, relevant_outputs):
        func_output = [bv.saved_output for bv in self.get_outputs()]
        jvp = None
        for f, x_dot in zip(self._vjp2funcs, tlm_inputs):
            if x_dot is None:
                continue
            tmp = Enlist(f(*inputs, *func_output, x_dot, **self._kwargs))
            if jvp is None:
                jvp = tmp
            else:
                for i in range(len(jvp)):
                    jvp[i] = jvp[i] + tmp[i]
        return jvp

    def evaluate_tlm_component(
        self, inputs, tlm_inputs, block_variable, idx, prepared=None
    ):
        return prepared[idx]

    def prepare_evaluate_hessian(
        self, inputs, hessian_inputs, adj_inputs, relevant_dependencies
    ):
        nterms = len(self._vjpfuncs)
        hvp = [None] * nterms
        tlm_inputs = [dep.tlm_value for dep in self.get_dependencies()]

        for (i, j), f in zip(product(range(nterms), range(nterms)), self._hesfuncs):
            x_dot = tlm_inputs[i]
            if x_dot is None:
                continue
            tmp = f(*inputs, *adj_inputs, x_dot, **self._kwargs)
            if hvp[j] is None:
                hvp[j] = tmp
            else:
                hvp[j] = hvp[j] + tmp
        return hvp

    def evaluate_hessian_component(
        self,
        inputs,
        hessian_inputs,
        adj_inputs,
        block_variable,
        idx,
        relevant_dependencies,
        prepared=None,
    ):
        return prepared[idx]

    def _jac_iterator(self, inputs, mask):
        if self._out_pointwise:
            # Set up all standard arguments that can be looped through pointwise.
            n = self.get_outputs()[0].saved_output.shape[0]
            standard_args = []
            for i, (a, pw) in enumerate(zip(inputs, self._pointwise)):
                if pw:
                    if a.shape[0] not in (1, n):
                        raise ValueError(
                            f"Argument {i} is marked as pointwise but first axis doesn't match output ({a.shape[0]} != {n})"
                        )
                    s = np.reshape(a, (a.shape[0], 1, *a.shape[1:]))
                    s = np.broadcast_to(s, (n, *s.shape[1:]), subok=True)
                else:
                    s = np.broadcast_to(a, (n, *a.shape), subok=True)
                standard_args.append(s)

        # Compute Autograd Jacobian df_dy.
        for input_shape, pw, jac_func, mask_i in zip(
            self._arg_shapes, self._pointwise, self._tlm_matrix_funcs, mask
        ):
            if mask_i is None:
                yield None
                continue
            if pw:
                # Need to extract some shape information and then create the outputs with the correct shape.
                # The shapes should be the following:
                #   -     arg_i: (n, *input_shape)
                #   -    df_dyi: (n, *output_shape, *input_shape).
                point_rank = len(input_shape)
                df_dyi_shape = (n, 1) + self._out_shape + (1,) + input_shape
                v = (jac_func(*args, **self._kwargs) for args in zip(*standard_args))
                df_dyi = np.zeros(df_dyi_shape)
                for k, df_dyik in enumerate(v):
                    df_dyi[k, ...] = df_dyik
                df_dyi = np.squeeze(df_dyi, axis=(1, -point_rank - 1))
            else:
                df_dyi = jac_func(*inputs, **self._kwargs)
            yield (df_dyi)

    def evaluate_tlm_matrix_component(
        self, inputs, tlm_inputs, block_variable, idx, prepared=None
    ):
        """
        If df_dy is the Jacobian of the Autograd function, this will give
        the matrix product df_dx = df_dy * dy_dx, where df_dx is a list
        df_dx[i] = df_dxi and dy_dx is a list of lists such that dy_dx[i][j]
        is dyi_dxj.
        """
        # Just renaming.
        dy_dx = tlm_inputs

        # Get length of return value.
        for dyi_dx in dy_dx:
            if dyi_dx is not None:
                num_controls = len(dyi_dx)
                break
        else:
            return None
        rv = [None] * num_controls

        # Compute df_dx[j] = sum_f(df_dy[i] * dy_dx[i][j])
        n = self.get_outputs()[0].saved_output.shape[0]
        df_dy = self._jac_iterator(inputs, dy_dx)
        for df_dyi, dyi_dx, input_shape, pw in zip(
            df_dy, dy_dx, self._arg_shapes, self._pointwise
        ):
            if dyi_dx is None:
                continue
            for j in range(num_controls):
                if dyi_dx[j] is None:
                    continue
                # Multiply df_dyi (Jacobian) by dyi_dx[j] (the input Jacobian).
                if isinstance(dyi_dx[j], JacobianIdentity):
                    df_dxj = df_dyi
                else:
                    point_rank = len(input_shape)
                    if pw:
                        # The `extra` is e.g. the shape of the control xj.
                        # The shapes should be the following:
                        #   -    df_dyi: (n, *output_shape, *input_shape).
                        #   - dyi_dx[j]: (n, *input_shape, *extra)
                        #   -    df_dxj: (n, *output_shape, *extra).
                        # df_dxj is computed by contracting `df_dyi` with dyi_dx[j]
                        # pointwise along the *input_shape axes.
                        extra = dyi_dx[j].shape[1 + point_rank :]
                        df_dxj_shape = (n,) + self._out_shape + extra
                        w = (
                            np.tensordot(a, b, axes=point_rank)
                            for a, b in zip(df_dyi, dyi_dx[j])
                        )
                        df_dxj = np.zeros(df_dxj_shape)
                        for k, df_dxjk in enumerate(w):
                            df_dxj[k, ...] = df_dxjk
                    else:
                        df_dxj = np.tensordot(df_dyi, dyi_dx[j], axes=point_rank)

                if rv[j] is None:
                    rv[j] = df_dxj
                else:
                    rv[j] += df_dxj
        return rv



from ufl import (
    FiniteElement,
    VectorElement,
    TensorElement,
    FunctionSpace,
    Coefficient,
    Argument,
    replace,
    inner,
)
from ufl.algorithms.analysis import extract_coefficients
from ufl.core.expr import Expr
from pyadjoint.enlisting import Enlist

from .types import PointMap, Space
from .space_builders import DirectSum, enlist


class UFLFunctionSpace(Space):
    """Represents a UFL FunctionSpace."""

    def __init__(self, functionspace):
        """
        Args:
            functionspace (ufl.FunctionSpace): the function space to wrap.
        """
        self._functionspace = functionspace

    def shape(self):
        return self._functionspace.ufl_element().value_shape()

    def is_point(self, point):
        return (
            hasattr(point, "ufl_function_space")
            and point.ufl_function_space() == self._functionspace
        )

    def point(self):
        return Coefficient(self._functionspace)

    def __eq__(self, other):
        return (
            isinstance(other, UFLFunctionSpace)
            and self._functionspace == other._functionspace
        )


class UFLExprSpace(Space):
    """Represents a space of UFL expression of a certain shape, defined by an
    example expression. Any UFL expression of the same shape lies in this space.
    """

    def __init__(self, expr, tlm_shape=None, ufl_domains=None):
        """
        Args:
            expr (ufl.core.expr.Expr): a UFL expression defining the space
            tlm_shape (tuple, optional): used to explicitly specify the shape of the space instead
                of getting it from the expression.

        TODO:
            * Rename ``tlm_shape`` arg to just ``shape``.
            * ufl_domains documentation
        """
        self._expr = expr
        self.ufl_shape = tlm_shape if tlm_shape is not None else expr.ufl_shape
        self.ufl_domains = (
            ufl_domains if ufl_domains is not None else expr.ufl_domains()
        )

    def shape(self):
        return self.ufl_shape

    def __repr__(self):
        domain_names = tuple(str(d) for d in self.ufl_domains)
        return f"UFLExprSpace(" + str(domain_names) + f",{self.ufl_shape!s})"

    def is_point(self, point):
        right_shape = isinstance(point, Expr) and point.ufl_shape == self.ufl_shape
        pd = point.ufl_domain()
        in_domain = self.ufl_domains is () or pd is None or (pd in self.ufl_domains)
        return right_shape and in_domain

    def point(self, domain_idx=None):
        """Returns the expression that this space was initialized with."""
        domain = None if domain_idx is None else self.ufl_domains[domain_idx]
        return Coefficient(
            FunctionSpace(domain, TensorElement("Real", shape=self.ufl_shape))
        )

    def __eq__(self, other):
        # I commented out the ufl_domains check since it can break the use case of
        #  converting from a domain-free space to a specific domain space, even though
        #  anything in the former can be in the latter.
        # To support that properly, we either need the idea of super/subspace comparisons
        #  or UFLtoUFL coverings that just use IdentityPointMaps.
        return isinstance(other, UFLExprSpace) and self.ufl_shape == other.ufl_shape
        # and self.ufl_domains == other.ufl_domains


class CR_UFL_Expr(PointMap):

    _abstract_element = FiniteElement("Real")

    def __init__(self, arg_space, exprs, pos_map, domain=None):
        """Take an expression for each component of the CR.

        Provide map for how the expression coefficients map to CR arguments."""

        # TODO: Entuple would be good instead of Enlist.
        self.exprs = Enlist(exprs)
        domains = [e.ufl_domain() for e in self.exprs]
        self.pos_map = pos_map

        ufl_domains = () if domain is None else (domain,)
        out_spaces = tuple(UFLExprSpace(e, ufl_domains=ufl_domains) for e in self.exprs)
        out_space = self.exprs.delist(DirectSum(*out_spaces))
        super(CR_UFL_Expr, self).__init__(arg_space, out_space)

    def __call__(self, args):
        if not isinstance(self.source, DirectSum):
            args = [args]
        self.inputs = args
        arg_map = {}
        out_list = []
        for key in self.pos_map:
            arg_map[key] = self.inputs[self.pos_map[key]]
        for expr in self.exprs:
            out_list.append(replace(expr, arg_map))
        return self.exprs.delist(out_list)

    def update_coefficients(self, replace_map):
        exprs = tuple(map(lambda e: replace(e, replace_map), self.exprs))

        # A hacky way to keep the enlistment the same.
        listed = self.exprs.listed
        self.exprs = Enlist(exprs)
        self.exprs.listed = listed


# Example
class CR_P_Laplacian(CR_UFL_Expr):
    """This CR is applied directly to ufl expressions"""

    def __init__(self, p=2, dim=2, input_u=True, domain=None):

        self._p = p
        self._input_u = input_u
        base_element = CR_UFL_Expr._abstract_element

        g = Coefficient(FunctionSpace(domain, VectorElement(base_element, dim=dim)))
        expr = g * (inner(g, g) + 1e-12) ** ((self._p - 2) / 2.0)

        if self._input_u:
            u = Coefficient(FunctionSpace(domain, base_element))
            arg_space = DirectSum(UFLExprSpace(u), UFLExprSpace(g))
            pos_map = {g: 1}
        else:
            arg_space = UFLExprSpace(g)
            pos_map = {g: 0}

        super(CR_P_Laplacian, self).__init__(arg_space, expr, pos_map)
        self.setParams(self._p)

    def setParams(self, p):
        self.update_coefficients({self._p: p})
        self._p = p
        self.params = self._p


def create_ufl_standins(shapes, domain=None):
    """Create standins for the outputs of a CR that can go in forms"""
    standins = []
    for s in shapes:
        if s:
            standins.append(
                Coefficient(FunctionSpace(domain, TensorElement("Real", shape=s)))
            )
        else:
            standins.append(Coefficient(FunctionSpace(domain, FiniteElement("Real"))))
    return tuple(standins)


def create_ufl_standin_arguments(shapes, number=0, domain=None):
    """Create standins for the outputs of a CR that can go in forms"""
    return tuple(
        Argument(FunctionSpace(domain, TensorElement("Real", shape=s)), number)
        for s in shapes
    )


def point_map(source_tuple, bare=False, domain=None, **kwargs):
    """This decorator turns the decorated function into a :class:`~crikit.cr.ufl.CR_UFL_Expr`.

    The given tuples are used to create :class:`~crikit.cr.ufl.UFLExprSpace` spaces to set as the
    source and target spaces.

    The decorated function will be run once to create the UFL expression for the point map.
    """
    # Get source_tuple in the form of a tuple of shapes.
    if source_tuple is None:
        source_tuple = ()
    elif not any(isinstance(s, tuple) for s in source_tuple):
        source_tuple = (source_tuple,)

    ufl_domains = () if domain is None else (domain,)

    # Generate inputs for the function and build the source space.
    inputs = create_ufl_standins(source_tuple)
    input_dict = {arg: i for i, arg in enumerate(inputs)}
    if len(inputs) == 0:
        source = None
    elif len(inputs) == 1:
        if bare:
            raise ValueError("bare should be False for functions with a single input")
        inputs = inputs[0]
        source = UFLExprSpace(inputs, ufl_domains=ufl_domains)
    else:
        source = DirectSum(
            tuple(UFLExprSpace(e, ufl_domains=ufl_domains) for e in inputs)
        )

    def point_map_decorator(ufl_func):
        outputs = ufl_func(*inputs) if bare else ufl_func(inputs)
        pm = CR_UFL_Expr(source, outputs, input_dict, domain=domain)
        return pm

    return point_map_decorator



from pyadjoint import *
from pyadjoint.enlisting import Enlist
from .adjfloat import AdjFloat
from .block import Block
from .block_variable import BlockVariable
from .identity import JacobianIdentity, make_jacobian_identities
from .tape import Tape, push_tape
from .reduced_function import ReducedFunction
from .reduced_function_numpy import ReducedFunctionNumPy
from .drivers import (
    compute_gradient,
    compute_jacobian_action,
    compute_jacobian_matrix,
    compute_hessian_action,
)
from .control import Control
from .solving.equation import ReducedEquation
from .solving.snes_solver import SNESSolver
from .verification import taylor_test, taylor_to_dict
from .tape_block import record_tape_block
from .minimize import minimize
from .callback import Callback, FileLoggerCallback, CallbackCombiner

try:
    from .numpy_adjoint import overload_autograd
except:
    pass

try:
    from .tensorflow_adjoint import get_params_feed_dict, run_tensorflow_graph
except:
    pass

from .fenics_adjoint import (
    homogenize_bcs,
    assemble,
    AssembleBlock,
    function_get_local,
    function_set_local,
    backend,
    compat,
)
from .jax_adjoint import (
    ndarray,
    array,
    asarray,
    to_jax,
    to_adjfloat,
    overload_jax,
    set_default_dtype,
    get_default_dtype,
)

# Make sure OverloadedType uses our version of BlockVariable.
from pyadjoint import OverloadedType


def create_block_variable(self):
    self.block_variable = BlockVariable(self)
    return self.block_variable


OverloadedType.create_block_variable = create_block_variable
del create_block_variable

from pyadjoint import set_working_tape

set_working_tape(Tape())



from ufl import Form
from fenics_adjoint import backend, Function
from fenics_adjoint.assembly import AssembleBlock as AssembleBlockOrig
from pyadjoint.tape import annotate_tape, stop_annotating, get_working_tape
from pyadjoint.overloaded_type import OverloadedType, create_overloaded_object

import backend
from fenics_adjoint.compat import compat

compat = compat(backend)


class OverloadedUFLArgument(object):
    """An overloaded argument whose adjoint values on a tape should be returned as rank-1 Forms"""

    pass


def assemble(*args, **kwargs):
    """Use the assemble syntax of firedrake, where the 'tensor' kwarg can take a Function.
    If that is the case (or if firedrake is the backend and the returned tensor is a Function),
    convert the result to an overloaded function so that an AssembleBlock can be created,
    even if the output is not a scalar."""

    annotate = annotate_tape(kwargs)

    with stop_annotating():
        if "tensor" in kwargs and isinstance(kwargs["tensor"], backend.Function):
            outputfunc = kwargs["tensor"]
            new_kwargs = kwargs.copy()
            new_kwargs["tensor"] = outputfunc.vector()
            backend.assemble(*args, **new_kwargs)
            output = outputfunc
        else:
            output = backend.assemble(*args, **kwargs)

    form = args[0]
    if isinstance(output, float) or isinstance(output, backend.Function):
        output = create_overloaded_object(output)

        if annotate:
            block = AssembleBlock(form)

            tape = get_working_tape()
            tape.add_block(block)

            block.add_output(output.block_variable)
    else:
        # Assembled a matrix
        output.form = form

    return output


class AssembleBlock(AssembleBlockOrig):
    def __init__(self, form):
        super(AssembleBlockOrig, self).__init__(form)
        self.form = form
        if backend.__name__ != "firedrake":
            mesh = self.form.ufl_domain().ufl_cargo()
        else:
            mesh = self.form.ufl_domain()
        self.add_dependency(mesh)
        for c in self.form.coefficients():
            if isinstance(c, OverloadedType):
                self.add_dependency(c, no_duplicates=True)

    def evaluate_adj_component(
        self, inputs, adj_inputs, block_variable, idx, prepared=None
    ):
        """catch when the adj_input is a Function, or when the return value
        should be a Form, otherwise do the default"""

        form = prepared
        adj_input = adj_inputs[0]
        c = block_variable.output
        c_rep = block_variable.saved_output

        args = form.arguments()
        n_args = len(args)
        if isinstance(c, OverloadedUFLArgument):
            if n_args == 0:
                form = adj_input * form
            elif n_args == 1:
                form = form * adj_input
            if isinstance(c, backend.Function):
                V = c.function_space()
            else:
                mesh = compat.extract_mesh_from_form(self.form)
                V = c._ad_function_space(mesh)
            dc = backend.TestFunction(V)
            dform = backend.derivative(form, c_rep, dc)
            return dform

        if n_args == 0:
            return super(AssembleBlock, self).evaluate_adj_component(
                inputs, adj_inputs, block_variable, idx, prepared=form
            )

        # In this case, adj_input may be a vector. It needs to be made into a function to put it into the form.
        if isinstance(adj_input, backend.Function):
            adj_input_func = adj_input
        else:
            # import crikit.utils as utils; utils.debug(locals(), globals())
            adj_input_func = backend.Function(args[0].function_space())
            adj_input_func.vector()[:] = adj_input[:]

        # if the output of the original is an assembled function, then
        # it should have exactly one argument, so we can apply it to adj_input.
        form = form * adj_input_func
        if isinstance(c, compat.ExpressionType):
            # Create a FunctionSpace from self.form and Expression.
            # And then make a TestFunction from this space.
            mesh = self.form.ufl_domain().ufl_cargo()
            V = c._ad_function_space(mesh)
            dc = backend.TestFunction(V)

            dform = backend.derivative(form, c_rep, dc)
            # print(c)
            # import crikit.utils as utils; utils.debug(locals(), globals())
            # output = backend.Function(c.function_space())
            # backend.assemble(dform,tensor=output.vector())
            output = compat.assemble_adjoint_value(dform)
            return [[output, V]]
        elif isinstance(c, compat.MeshType):
            X = backend.SpatialCoordinate(c_rep)
            dform = backend.derivative(form, X)
            output = compat.assemble_adjoint_value(dform)
            return output
        if isinstance(c, backend.Function):
            V = c.function_space()
        else:
            mesh = compat.extract_mesh_from_form(self.form)
            V = c._ad_function_space(mesh)
        dc = backend.TestFunction(V)
        dform = backend.derivative(form, c_rep, dc)
        output = compat.assemble_adjoint_value(dform)
        return output

    def evaluate_tlm_component(
        self, inputs, tlm_inputs, block_variable, idx, prepared=None
    ):
        """catch when the output is a Function, otherwise do the default"""
        output = super(AssembleBlock, self).evaluate_tlm_component(
            inputs, tlm_inputs, block_variable, idx, prepared=prepared
        )
        saved_output = self.get_outputs()[idx].saved_output
        if isinstance(saved_output, backend.Function):
            output = backend.Function(saved_output.function_space(), output)
        return output

    def evaluate_hessian_component(
        self,
        inputs,
        hessian_inputs,
        adj_inputs,
        block_variable,
        idx,
        relevant_dependencies,
        prepared=None,
    ):
        form = prepared
        hessian_input = hessian_inputs[0]
        adj_input = adj_inputs[0]

        c1 = block_variable.output
        c1_rep = block_variable.saved_output

        args = form.arguments()
        n_args = len(args)
        if n_args > 0 or isinstance(
            c1, OverloadedUFLArgument
        ):  # or isinstance(adj_input, backend.Function):
            if isinstance(adj_input, backend.Function):
                adj_input_func = adj_input
            else:
                adj_input_func = backend.Function(args[0].function_space())
                adj_input_func.vector()[:] = adj_input[:]

            if isinstance(hessian_input, backend.Function):
                hessian_input_func = hessian_input
            else:
                hessian_input_func = backend.Function(args[0].function_space())
                hessian_input_func.vector()[:] = hessian_input[:]

            if n_args == 0:
                hform = adj_input * form
                aform = hessian_input * form
            else:
                hform = form * hessian_input_func
                aform = form * adj_input_func
            mesh = self.form.ufl_domain().ufl_cargo()
            if isinstance(c1, backend.Function):
                V = c1.function_space()
            else:
                V = c1._ad_function_space(mesh)
            dc = backend.TestFunction(V)
            dform = backend.derivative(hform, c1_rep, dc)
            dforma = backend.derivative(aform, c1_rep, dc)
            for other_idx, bv in relevant_dependencies:
                c2_rep = bv.saved_output
                tlm_input = bv.tlm_value

                if tlm_input is None:
                    continue

                if isinstance(c2_rep, compat.MeshType):
                    X = backend.SpatialCoordinate(c2_rep)
                    ddform = backend.derivative(dforma, X, tlm_input)
                else:
                    ddform = backend.derivative(dforma, c2_rep, tlm_input)
                dform += ddform
            if not isinstance(c1, OverloadedUFLArgument):
                return compat.assemble_adjoint_value(dform)
            return dform

        return super(AssembleBlock, self).evaluate_hessian_component(
            inputs,
            hessian_inputs,
            adj_inputs,
            block_variable,
            idx,
            relevant_dependencies,
            prepared,
        )

    def recompute_component(self, inputs, block_variable, idx, prepared):
        saved_output = self.get_outputs()[idx].saved_output
        if isinstance(saved_output, backend.Function):
            form = prepared
            V = saved_output.function_space()
            output = backend.Function(V)
            backend.assemble(form, tensor=output.vector())
            output = create_overloaded_object(output)
            return output
        else:
            return super(AssembleBlock, self).recompute_component(
                inputs, block_variable, idx, prepared
            )



from pyadjoint_utils.numpy_adjoint import ndarray
from pyadjoint_utils.block import Block
from pyadjoint_utils.identity import JacobianIdentity
from pyadjoint.enlisting import Enlist
from pyadjoint.tape import get_working_tape, stop_annotating, annotate_tape
from pyadjoint.overloaded_type import create_overloaded_object
from functools import wraps

def _array(x):
    return ndarray._ad_init_object(x)

class ONNXBlock(Block):

    def __init__(self,
                 model,
                 jacobian,
                 args,
                 outputs
    ):

        super().__init__()
        # protect the import in case we don't have onnxruntime installed
        # and don't want to use this class
        import onnxruntime as ort
        for arg in args:
            self.add_dependency(arg)
        for out in outputs:
            self.add_output(out.create_block_variable())

        self.model = self._load_model(model) if isinstance(model, str) else model
        self.jacobian = self._load_model(jacobian) if isinstance(jacobian, str) else jacobian

        self._args = list(args)
        self._argnums = list(range(len(args)))

    def _replace_params_partial(self, new_params, idxs):
        for i, idx in enumerate(idxs):
            self._args[idx] = new_params[i]

    def _make_partial_func(self, idxs, f):
        def _pdiff_func(*args):
            param_idxs = [self._argnums[idx] for idx in idxs]
            self._replace_params_partial(args, param_idxs)
            return f(*self._args)

        return _pdiff_func

    def _load_model(self, filename):
        session = ort.InferenceSession(filename)
        inames = [node.name for node in session.get_inputs()]

        def model(*args):
            if len(args) != len(inames):
                raise ValueError("Must provide the same number of args as input names ({len(inames)}) to the model!")
            return tuple(map(_array, session.run(None, {name : arg for name, arg in zip(inames, args)})))

        return model


    def prepare_recompute_component(self, inputs, relevant_outputs):
        return Enlist(self.model(*inputs))
    

    def recompute_component(self, inputs, block_variable, idx, prepared):
        return prepared[idx]


    def prepare_evaluate_tlm_matrix(self, inputs, tlm_inputs, relevant_outputs):
        args = []
        argnums = []
        for i, x in enumerate(tlm_inputs):
            if inputs[i] is not None:
                args.append(inputs[i])
                argnums.append(i)
            if x is not None:
                for j, di_dj in enumerate(x):
                    if (i != j and di_dj is not None) or (
                        i == j and not isinstance(di_dj, JacobianIdentity)
                    ):
                        raise NotImplementedError(
                            "Non-identity tlm_inputs cannot be handled yet!"
                        )

        jacfun = self._make_partial_func(argnums, self.jacobian)
        return jacfun(*args)


    def evaluate_tlm_matrix_component(
        self, inputs, tlm_inputs, block_variable, idx, prepared=None
    ):
        return prepared[idx]




def overload_onnx(model, jacobian=None):

    @wraps(model)
    def _overloaded_model(*args):
        with stop_annotating():
            out = Enlist(model(*args))

        overloads = [create_overloaded_object(x) for x in out]
        if annotate_tape():
            get_working_tape().add_block(
                ONNXBlock(model, jacobian, args, overloads)
            )

        return out.delist(overloads)

    return _overloaded_model
        



import time
import datetime
import platform
import getpass
import logging
from typing import Sequence, Callable
from .reduced_function import ReducedFunction
from .reduced_function_numpy import ReducedFunctionNumPy
from pyadjoint.reduced_functional import ReducedFunctional
from pyadjoint.reduced_functional_numpy import ReducedFunctionalNumPy

logger = logging.getLogger("CRIKit")

try:
    from mpi4py import MPI
except ImportError:
    # if we can't import mpi4py, then CRIKit is installed without
    # MPI, meaning everything is serial
    class Comm:
        def Get_rank(self):
            return 0

        def Get_size(self):
            return 1

    class MPI:

        COMM_WORLD = Comm()
        COMM_SELF = Comm()


class Callback:
    """Base class for CRIKit callbacks.

    :param name: The name of the callback, defaults to 'Callback'
    :type name: str
    """

    name: str = "Callback"

    def __init__(self, name=None):
        if name is not None:
            self.name = name

    def __call__(self, xk, state=None):
        """Evaluates the callback at a given point with given state"""
        raise NotImplementedError


def now():
    return datetime.datetime.now().ctime()


class FileLoggerCallback(Callback):
    """A callback that logs info about the problem being solved to a file.

    :param name: The name of this callback. Useful for determining which
    of several callbacks wrote to a given file. Defaults to
    `'minimization_callback'`.
    :type name: str
    :param problem_type: A string describing what sort of problem is being
    solved. Defaults to `'Minimization'`
    :type problem_type: str
    :param filename: The filename of a file to log information about this
    run (including machine and timing info) to. Defaults to
    `'{name}_info.txt'` where `name` is the name of this callback.
    :type filename: str, optional
    :param overwrite_file: If `True`, open the file specified by `filename`
    in write mode instead of append mode, deleting the current contents of
    the file before writing to it. Defaults to False
    :type overwrite_file: bool
    :param rf: The `ReducedFunctional` or `ReducedFunction` used in the
    problem being solved. We use this to get the current values of the
    parameters for said `rf`.
    :param write_params: If `True` and `rf` is not `None`, write the
    parameters in the log. Defaults to True
    :type write_params: bool
    :param mpi_comm: The name of the active MPI communicator, which is
    retrieved via `getattr(mpi4py.MPI, mpi_comm)`, with sensible defaults
    if `MPI` is not installed. This is needed if and only if
    `info_file` is not None, and exists to ensure the file is written only
    on MPI rank 0. Defaults to `COMM_WORLD`.
    :type mpi_comm: str
    """

    def __init__(
        self,
        name="minimize_callback",
        problem_type="Minimization",
        filename=None,
        overwrite_file=False,
        rf=None,
        write_params=True,
        mpi_comm="COMM_WORLD",
    ):

        super().__init__(name=name)
        self.filename = self.name + "_info.txt" if filename is None else filename
        self.problem_type = problem_type
        self.rf = rf
        self.write_params = write_params
        self.mpi_comm = getattr(MPI, mpi_comm)
        self.num_calls = 0
        self.mpi_rank = self.mpi_comm.Get_rank()
        if self.mpi_rank == 0:
            with open(self.filename, "w" if overwrite_file else "a") as f:
                f.write(self._build_header())

    def _build_header(self):
        kernel, node, release, os_version, machine, processor = platform.uname()
        header = (
            self.problem_type
            + " Problem "
            + self.name
            + " on platform "
            + f"{kernel}-{release} on node {node} with OS version {os_version} by user {getpass.getuser()} "
            + f"and machine type {machine}"
        )
        if machine != processor:
            header += " and processor {processor}"

        header += "\nStarting at " + now() + "\n"
        return header

    def _format_params(self, xk):
        if isinstance(self.rf, ReducedFunctional):
            cls = ReducedFunctionalNumPy
        else:
            # if rf is not a ReducedFunctional, it's a ReducedFunction
            cls = ReducedFunctionNumPy

        return str(
            cls(self.rf).set_local([c.tape_value() for c in self.rf.controls], xk)
        )

    def _build_message(self, xk):
        if self.write_params and self.rf is not None:
            params = self._format_params(xk)
            with_params = " with parameters "
        else:
            params = with_params = ""

        self.num_calls += 1
        return (
            now()
            + " : Iteration number "
            + str(self.num_calls)
            + with_params
            + params
            + "\n"
        )

    def __call__(self, xk, state=None):
        if self.mpi_rank == 0:
            with open(self.filename, "a") as f:
                f.write(self._build_message(xk))


class CallbackCombiner(Callback):
    """A {class}`CallbackCombiner` takes in several callback functions
    or classes (which must provide a `__call__` method) and calls all of them
    in order, each inside its own try-catch block. Exceptions are reported
    through the `logging` module at level `ERROR`.

    :param callbacks: A list or tuple of callback functions/classes to callback
    :type callbacks: Sequence[Callable]
    :param name: The name of this instance, defaults to `'CallbackCombiner'`
    :type name: str
    """

    def __init__(self, callbacks, name="CallbackCombiner"):
        self.callbacks = callbacks
        super().__init__(name=name)

    def __call__(self, *args, **kwargs):
        """Evalute the callbacks in order with given arguments. Exceptions
        are reported through the `logging` module at log level `ERROR`.
        """
        for cb in self.callbacks:
            try:
                cb(*args, **kwargs)
            except Exception as e:
                logging.error("Caught exception in " + self.name + ":" + str(e))



from ..block import Block
from pyadjoint.overloaded_type import OverloadedType
from ..adjfloat import AdjFloat
from pyadjoint.enlisting import Enlist
import jax
import jax.numpy as jnp
from jax.tree_util import Partial as partial


class ArrayOperatorBlock(Block):
    # a simpler JAXBlock that's independent of the actual JAXBlock implementation so as not to create a circular dependency
    def __init__(self, operator, args, output):
        super().__init__()
        self._operator = operator
        self._args = [convert_arg(x) for x in args]
        for dep in args:
            self.add_dependency(dep)

    def recompute_component(self, inputs, block_variable, idx, prepared):
        return self._operator(*(convert_arg(x) for x in inputs))


class AddBlock(ArrayOperatorBlock):
    def evaluate_adj_component(
        self, inputs, adj_inputs, block_variable, idx, prepared=None
    ):
        return adj_inputs[0]

    def evaluate_tlm_component(
        self, inputs, tlm_inputs, block_variable, idx, prepared=None
    ):
        x_dot, y_dot = (convert_arg(val) for val in tlm_inputs)
        return sum([x if x is not None else 0 for x in [x_dot, y_dot]])

    def evaluate_hessian_component(
        self,
        inputs,
        hessian_inputs,
        adj_inputs,
        block_variable,
        idx,
        relevant_dependencies,
        prepared=None,
    ):
        return hessian_inputs[0]


class SubBlock(ArrayOperatorBlock):
    def evaluate_adj_component(
        self, inputs, adj_inputs, block_variable, idx, prepared=None
    ):
        if idx == 0:
            return adj_inputs[0]
        return -adj_inputs[0]

    def evaluate_tlm_component(
        self, inputs, tlm_inputs, block_variable, idx, prepared=None
    ):
        x_dot, y_dot = (convert_arg(val) for val in tlm_inputs)
        output = 0
        if x_dot is not None:
            output = output + x_dot
        if y_dot is not None:
            output = output - y_dot

        return output


class MulBlock(ArrayOperatorBlock):
    def evaluate_adj_component(
        self, inputs, adj_inputs, block_variable, idx, prepared=None
    ):
        other_idx = 0 if idx == 1 else 1
        return self._operator(
            convert_arg(adj_inputs[0]), convert_arg(inputs[other_idx])
        )

    def evaluate_tlm_component(
        self, inputs, tlm_inputs, block_variable, idx, prepared=None
    ):
        x_dot, y_dot = (convert_arg(val) for val in tlm_inputs)
        output = 0
        if x_dot is not None:
            output = output + self._operator(convert_arg(inputs[1]), x_dot)
        if y_dot is not None:
            output = output + self._operator(convert_arg(inputs[0]), y_dot)
        return output


@jax.jit
def _compiled_pow(x, y):
    # for use in DivBlock
    return x ** y


class DivBlock(ArrayOperatorBlock):
    def evaluate_adj_component(
        self, inputs, adj_inputs, block_variable, idx, prepared=None
    ):
        z_bar = convert_arg(adj_inputs[0])
        if idx == 0:
            return z_bar / convert_arg(inputs[1])
        else:
            return (
                -z_bar
                * convert_arg(inputs[0])
                / _compiled_pow(convert_arg(inputs[1]), 2)
            )

    def evaluate_tlm_component(
        self, inputs, tlm_inputs, block_variable, idx, prepared=None
    ):

        x, y = (convert_arg(val) for val in inputs)
        x_dot, y_dot = (convert_arg(val) for val in tlm_inputs)

        @jax.jit
        def _tlm_arg_0(x, y, x_dot):
            return x_dot / y

        @jax.jit
        def _tlm_arg_1(x, y, y_dot):
            return -y_dot * (x / (y ** 2))

        tlm = []

        if x_dot is not None:
            tlm.append(_tlm_arg_0(x, y, x_dot))
        if y_dot is not None:
            tlm.append(_tlm_arg_1(x, y, y_dot))

        return sum(tlm)


class NegBlock(ArrayOperatorBlock):
    def evaluate_adj_component(
        self, inputs, adj_inputs, block_variable, idx, prepared=None
    ):
        return -convert_arg(adj_inputs[0])

    def evaluate_tlm_component(
        self, inputs, tlm_inputs, block_variable, idx, prepared=None
    ):
        return -convert_arg(tlm_inputs[0])


class PowBlock(ArrayOperatorBlock):
    def evaluate_adj_component(
        self, inputs, adj_inputs, block_variable, idx, prepared=None
    ):
        x, y = (convert_arg(val) for val in inputs)
        z_bar = convert_arg(adj_inputs[0])

        @jax.jit
        def _arg_0_vjp(x, y, z_bar):
            # vjp w.r.t. x; derivative is y * x ** (y - 1)
            return z_bar * y * (x ** (y - 1))

        @jax.jit
        def _arg_1_vjp(x, y, z_bar):
            # vjp w.r.t. y; derivative is log(x) * x ** y
            return z_bar * jnp.log(x) * (x ** y)

        if idx == 0:
            return _arg_0_vjp(x, y, z_bar)
        else:
            return _arg_1_vjp(x, y, z_bar)

    def evaluate_tlm_component(
        self, inputs, tlm_inputs, block_variable, idx, prepared=None
    ):

        x, y = (convert_arg(val) for val in inputs)
        x_dot, y_dot = (convert_arg(val) for val in tlm_inputs)

        @jax.jit
        def _tlm_arg_0(x, y, x_dot):
            return x_dot * y * (x ** (y - 1))

        @jax.jit
        def _tlm_arg_1(x, y, y_dot):
            return y_dot * jnp.log(x) * (x ** y)

        tlm = []

        if x_dot is not None:
            tlm.append(_tlm_arg_0(x, y, x_dot))
        if y_dot is not None:
            tlm.append(_tlm_arg_1(x, y, y_dot))
        return sum(tlm)


def convert_arg(x):
    if not isinstance(x, OverloadedType):
        return x

    if hasattr(x, "value"):
        return x.value
    elif hasattr(x, "saved_output"):
        return convert_arg(x.saved_output)
    else:
        return x  # see if it works! (it probably won't, and you'll get an exception from a function higher up the call chain)



import jax.numpy as np
import numpy as onp
import jax

from crikit.covering import Covering, register_covering
from crikit.covering.ufl import get_numpy_shape
from crikit.cr.types import PointMap, Space
from crikit.cr.ufl import UFLExprSpace, UFLFunctionSpace
from crikit.cr.quadrature import get_quadrature_params, make_quadrature_spaces
from crikit.fe import *
from crikit.fe_adjoint import *
from crikit.projection import project
from pyadjoint_utils.jax_adjoint import array, ndarray
from pyadjoint_utils.fenics_adjoint import function_get_local, function_set_local
from pyadjoint_utils import ReducedFunctionNumPy, ReducedFunction
from pyadjoint.tape import (
    no_annotations,
    annotate_tape,
    get_working_tape,
    stop_annotating,
)
from pyadjoint.enlisting import Enlist
from pyadjoint_utils.convert import make_convert_block
from pyadjoint.overloaded_function import overload_function


class JAXArrays(Space):
    """
    See Ndarrays in crikit/cr/numpy.py

    Args:
       shape (Iterable): the shape of the arrays
       dtype (jax.numpy.dtype): the data type (default None)
    """

    def __init__(self, shape, dtype=None):
        self._shape = tuple(shape)
        self._dtype = dtype
        self._indefinite_axes = np.array(shape) < 0
        self._definite_axes = np.logical_not(self._indefinite_axes)
        self._definite_shape = np.asarray(shape)[self._definite_axes]

    @property
    def shape(self):
        return self._shape

    def shape(self):
        return self._shape

    def is_point(self, point):
        if not isinstance(
            point, (np.ndarray, jax.interpreters.xla.DeviceArray, ndarray)
        ):
            return False
        if len(point.shape) != len(self._shape):
            return False
        if self._dtype is not None and self._dtype != point.dtype:
            return False
        return np.array_equal(
            np.asarray(point.shape)[self._definite_axes], self._definite_shape
        )

    def point(self, **kwargs):
        return np.zeros(self._shape)

    def __eq__(self, other):
        return (
            isinstance(other, JAXArrays)
            and self._shape == other._shape
            and self._dtype == other._dtype
        )

    def __repr__(self):
        if self._dtype is None:
            return f"JAXArrays({self._shape})"
        return f"JAXArrays({self._shape}, dtype={self._dtype})"


class JAX_To_UFLFunctionSpace(PointMap):
    def __init__(
        self, source, target, quad_space=None, quad_params=None, make_block=True
    ):
        if quad_space is None:
            quad_space, quad_params = make_quadrature_spaces(
                target, quad_params=quad_params
            )

        self._quad_space = quad_space
        self._dx = dx(metadata=quad_params)

        if isinstance(target, UFLFunctionSpace):
            self._target_space = target._functionspace
        else:
            self._target_space = quad_space

        if source is None:
            quad_shape = get_numpy_shape(self._quad_space)
            source = JAXArrays(quad_shape)

        super().__init__(source, target)

    def __call__(self, arr, **kwargs):
        q = function_set_local(Function(self._quad_space), arr)
        if self._target_space != self._quad_space:
            q = project(q, self._target_space, dx=self._dx)
        return q


class UFLExprSpace_To_JAX(PointMap):
    """Much like UFLExprSpace_To_Numpy, this class maps a UFL expression into a JAX array. The constructor inputs are the same here, replacing Ndarrays with JAXArrays
    Args:
        source (UFLExprSpace or UFLFunctionSpace): the UFL space to use as input.
        target (JAXArrays, optional): the target space to map to.
        quad_space (optional): the finite-element quadrature space to interpolate to.
        quad_params (dict, optional): parameters for the quadrature space.
        domain (optional): the UFL domain for the quadrature space.
    """

    def __init__(
        self, source, target=None, quad_space=None, quad_params=None, domain=None
    ):
        if quad_space is None:
            quad_space, quad_params = make_quadrature_spaces(
                source, quad_params=quad_params, domain=domain
            )
        self._quad_space = quad_space
        self._dx = dx(metadata=quad_params)

        if target is None:
            quad_shape = get_numpy_shape(self._quad_space)
            target = JAXArrays(quad_shape)

        super().__init__(source, target)

    def __call__(self, expr, **kwargs):
        if expr.ufl_shape != self._quad_space.ufl_element().value_shape():
            raise ValueError(
                f"Expression shape {expr.ufl_shape} does not match the target shape of {self._quad_space.ufl_element().value_shape()}"
            )
        if isinstance(expr, Function) and self._quad_space == expr.function_space():
            func = expr
        else:
            func = project(expr, self._quad_space, dx=self._dx)
        res = function_get_local(func)
        return convert_numpy_to_jax(res)


def backend_convert_numpy_to_jax(x):
    return np.array(x)


def backend_convert_jax_to_numpy(x):
    return onp.array(x)


ConvertNumpyToJAX = make_convert_block(
    backend_convert_numpy_to_jax, backend_convert_jax_to_numpy, "ConvertNumpyToJAX"
)
ConvertJAXToNumpy = make_convert_block(
    backend_convert_jax_to_numpy, backend_convert_numpy_to_jax, "ConvertJAXToNumpy"
)

convert_numpy_to_jax = overload_function(
    backend_convert_numpy_to_jax, ConvertNumpyToJAX
)
convert_jax_to_numpy = overload_function(
    backend_convert_jax_to_numpy, ConvertJAXToNumpy
)


@register_covering(UFLExprSpace, JAXArrays)
@register_covering(UFLFunctionSpace, JAXArrays)
class JAX_UFLFunctionSpace_Covering(Covering):
    def __init__(
        self, base_space, covering_space=None, domain=None, quad_params=None, **kwargs
    ):
        quad_space, quad_params = make_quadrature_spaces(
            base_space, quad_params=quad_params, domain=domain
        )
        self._quad_space = quad_space
        self._quad_params = quad_params

        if covering_space is None:
            quad_shape = get_numpy_shape(self._quad_space)
            covering_space = JAXArrays(quad_shape)

        super().__init__(base_space, covering_space, **kwargs)

    def covering_map(self):
        return JAX_To_UFLFunctionSpace(
            self._covering_space,
            self._base_space,
            quad_space=self._quad_space,
            quad_params=self._quad_params,
        )

    def section_map(self):
        return UFLExprSpace_To_JAX(
            self._base_space,
            self._covering_space,
            quad_space=self._quad_space,
            quad_params=self._quad_params,
        )


class ReducedFunctionJAX(ReducedFunctionNumPy):
    """This class implements the ReducedFunction for a given ReducedFunction
    and controls with JAX data structures. Like a ReducedFunctionNumPy, these
    are created from ReducedFunction instances like
    >>> from pyadjoint_utils import overload_jax
    >>> f = overload_jax(lambda x: np.sum(x ** 2))
    >>> x = array(np.array([1.0,2.0]))
    >>> rf = ReducedFunction(f(x),Control(x))
    >>> rf_jax = ReducedFunctionJAX(rf)
    >>> float(rf_jax(x))
    5.0
    """

    def __init__(self, rf):
        if not isinstance(rf, ReducedFunction):
            raise TypeError(
                f"Must pass a ReducedFunction to the ReducedFunctionJAX constructor, not a {type(rf)}!"
            )

        self.rf = rf

    def __call__(self, val):
        ip = self.get_rf_input(array(val))
        output = self.rf(ip)
        return output

    def get_global(self, controls):
        ctrls = []
        for i, val in enumerate(controls):
            if isinstance(val, Control):
                ctrls += val.fetch_numpy(val.control)
            elif hasattr(val, "_ad_to_list"):
                ctrls += val._ad_to_list(val)
            else:
                ctrls += self.controls[i].control._ad_to_list(val)

        return ndarray(np.array(ctrls))

    @no_annotations
    def jac_action(self, val):
        ip = self.get_rf_input(val)
        dJdp = self.rf.jac_action(ip)
        return self.get_outputs_array(dJdp)

    def get_outputs_array(self, vals):
        outs = []
        vals = Enlist(vals)
        for i, out in enumerate(self.outputs):
            if vals[i] is not None:
                outs += out.output._ad_to_list(vals[i])
            else:
                outs += [0] * out.output._ad_dim()

        return ndarray(np.array(outs))



from typing import Any, Tuple


Shape = Tuple[int, ...]


class Space:
    """Arguments / output spaces of CRs.

    A class of spaces is a representation.

    Spaces in a representation may look something like expression
    trees of operations that build new spaces recursively from
    base spaces.

    Abstract representations will have spaces that can't instantiate points as
    real data.  They have to be lowered into concrete representations.
    """

    def point(self, *args, **kwargs):
        """create a point in the space"""
        raise NotImplementedError

    def is_point(self, x: Any) -> bool:
        """test if a point is in the space"""
        raise NotImplementedError

    def shape(self) -> Shape:
        """Get the shape of the space (as in the shape of the arguments
        that you would make from a point in the space)"""
        raise NotImplementedError


class PointMap:
    """Takes points from one space and maps them into another"""

    def __init__(self, source_space, target_space):
        self._source = source_space
        self._target = target_space

    @property
    def source(self):
        """the input Space of the PointMap"""
        return self._source

    @property
    def target(self):
        """the output Space of the PointMap"""
        return self._target

    def __call__(self, point, **kwargs):
        """Apply the point map to the given point. I.e., map the given point in
        the source space to a point in the target space."""
        raise NotImplementedError

    def est_degree(self, *args, **kwargs) -> int:
        return None

    def adjoint(self, point, adj_input):
        raise NotImplementedError

    def tlm(self, point, tlm_input):
        raise NotImplementedError

    def hessian(self, arg, adj_input, tlm_input):
        raise NotImplementedError



from pyadjoint.tape import (
    get_working_tape,
    set_working_tape,
    stop_annotating,
    annotate_tape,
)
from pyadjoint.block import Block
from pyadjoint.overloaded_type import create_overloaded_object
from pyadjoint_utils import *
import numpy as np
from fenics import Function, FunctionSpace, MixedElement


def backend_run_network(network, output_space, *args, **kwargs):
    # Each function is reshaped to have value_dimension() columns.
    inputs = []
    for f in args:
        v = np.array(f.vector()).reshape((-1, f.value_dimension(0)))
        inputs.append(v)

    # Concatenate all the input columns together so that each row has all the info for a single point.
    net_in = np.hstack(inputs)

    # Run the network.
    output = network.predict(net_in, **kwargs)

    # TODO: I'm assuming there's only one output for now.
    # Turn the network output into a Function in Q.
    func = Function(output_space)

    func.vector()[:] = output.flatten()

    return func


def run_network(network, output_space, *args, **kwargs):
    annotate = annotate_tape(kwargs)

    if annotate:
        tape = get_working_tape()
        network = create_overloaded_object(network)
        block = RunNetworkBlock(network, output_space, *args, **kwargs)
        tape.add_block(block)

    with stop_annotating():
        output = backend_run_network(network, output_space, *args, **kwargs)

    output = create_overloaded_object(output)

    if annotate:
        block.add_output(output.create_block_variable())

    return output


def get_num_network_inputs_outputs(output_space, *args):
    num_inputs = 0
    for f in args:
        if isinstance(f, Function):
            num_inputs += f.value_dimension(0)
        else:
            num_inputs += f.element().value_dimension(0)

    num_outputs = output_space.element().value_dimension(0)

    return num_inputs, num_outputs


class RunNetworkBlock(Block):
    def __init__(self, network, output_space, *args, **kwargs):
        super(RunNetworkBlock, self).__init__()
        self.add_dependency(network)
        for f in args:
            self.add_dependency(f)
        self.kwargs = kwargs
        self.output_space = output_space

    def __str__(self):
        return "RunNetworkBlock"

    def prepare_evaluate_adj(self, inputs, adj_inputs, relevant_dependencies):
        adj_input = adj_inputs[0]
        net = inputs[0]

        # Assume the network has only one output.
        # Get output into the right shape (column vector).
        output_shape = self.output_space.ufl_element().value_shape()
        if len(output_shape) == 0:
            output_shape = (1,)
        adj_input = np.array(adj_input).reshape((-1, *output_shape))

        net.derivative_calculations(adj_input, debug=False)

        return net.input_err

    def evaluate_adj_component(
        self, inputs, adj_inputs, block_variable, idx, prepared=None
    ):
        if idx == 0:
            net = block_variable.saved_output
            return net.create_copy(weights=net.param_err)

        # This is the idx-th function.
        func = block_variable.saved_output
        Q = func.function_space()

        # Grab the columns corresponding to this function.
        start = 0
        for i in range(1, idx):
            start += inputs[i].value_dimension(0)
        adj_output = prepared[:, start : start + inputs[idx].value_dimension(0)]

        # The adj_output should be in the same FunctionSpace as the corresponding input.
        q = Function(Q)
        q.vector()[:] = adj_output.flatten()
        return q.vector()

    def prepare_evaluate_tlm(self, inputs, tlm_inputs, relevant_outputs):
        net = inputs[0]
        net.jacobian_calculations()
        return (net.jacobian, net.p_jacobian)

    def evaluate_tlm_component(
        self, inputs, tlm_inputs, block_variable, idx, prepared=None
    ):
        net_jacobian, net_p_jacobian = prepared

        # Grab the columns corresponding to each function and multiply by the tlm_input for that function
        start = 0
        rv = None
        for i, tlm_input in enumerate(tlm_inputs):
            if tlm_input is None:
                continue
            if i == 0:
                df_dyi = net_p_jacobian
                if tlm_input is not None and df_dyi is None:
                    raise NotImplementedError(
                        "Jacobian calculations are not complete for %s"
                        % inputs[0].__class__.__name__
                    )
            else:
                # Extract the entries related to this input function.
                df_dyi = net_jacobian[:, start : start + inputs[i].value_dimension(0)]

            if isinstance(tlm_input, JacobianIdentity):
                dot = df_dyi
            else:
                # First, get tlm_input into a numpy array of the right shape.
                if hasattr(tlm_input, "vector"):
                    tlm_in_mat = tlm_input.vector()[:]
                elif hasattr(tlm_input, "values"):
                    tlm_in_mat = tlm_input.values()
                else:
                    raise NotImplementedError(
                        "Can only handle Constant and Function inputs"
                    )

                shape = list(tlm_input.ufl_shape)
                while len(shape) < 2:
                    shape.append(1)
                tlm_in_mat = tlm_in_mat.reshape(-1, *shape)

                # Now do the multication. In the general case, this should probably be a np.tensor dot,
                # but I'll get to that later, as needed.
                dot = np.matmul(df_dyi, tlm_in_mat)

            if rv is None:
                rv = dot
            else:
                rv += dot

            if i != 0:
                start += inputs[i].value_dimension(0)

        if rv is None:
            return
        r = Function(self.output_space)
        r.vector()[:] = rv.flatten()
        return r

    def prepare_evaluate_tlm_matrix(self, inputs, tlm_inputs, relevant_outputs):
        net = inputs[0]
        net.jacobian_calculations()
        return (net.jacobian, net.p_jacobian)

    def evaluate_tlm_matrix_component(
        self, inputs, tlm_inputs, block_variable, idx, prepared=None
    ):
        net_jacobian, net_p_jacobian = prepared

        for tlm_input in tlm_inputs:
            if tlm_input is not None:
                num_controls = len(tlm_input)
                break
        else:
            return None
        rv = [None] * num_controls

        # Grab the columns corresponding to each function and multiply by the tlm_input for that function
        output_shape = self.output_space.ufl_element().value_shape()
        start = 0
        for i, dyi_dx in enumerate(tlm_inputs):
            if dyi_dx is None:
                continue
            if i == 0:
                df_dyi = net_p_jacobian
                if dyi_dx is not None and df_dyi is None:
                    raise NotImplementedError(
                        "Jacobian calculations are not complete for %s"
                        % inputs[0].__class__.__name__
                    )
            else:
                # Extract the entries related to this input function.
                df_dyi = net_jacobian[:, start : start + inputs[i].value_dimension(0)]
            for j in range(num_controls):
                if dyi_dx[j] is None:
                    continue
                # Now I need to multiply df_dyi by dyi_dx[j].

                # First, get dyi_dx[j] into a numpy array of the right shape.
                if hasattr(dyi_dx[j], "vector"):
                    tlm_in_mat = dyi_dx[j].vector()[:]
                elif hasattr(dyi_dx[j], "values"):
                    tlm_in_mat = dyi_dx[j].values()
                else:
                    raise NotImplementedError(
                        "Can only handle Constant and Function inputs"
                    )
                tlm_in_mat = tlm_in_mat.reshape(-1, *dyi_dx[j].tlm_shape)

                # Now do the multication. In the general case, this should probably be a np.tensor dot,
                # but I'll get to that later, as needed.
                output = np.matmul(df_dyi, tlm_in_mat)

                if rv[j] is None:
                    rv[j] = output
                else:
                    rv[j] += output

            if i != 0:
                start += inputs[i].value_dimension(0)

        for j, tlm_output in enumerate(rv):
            if tlm_output is None:
                continue
            # I'm representing the Jacobian as a MixedElement where each mixed element consists of the
            # output element repeated multiple times.
            output_shape_indices = len(self.output_space.ufl_element().value_shape())
            tlm_shape = tlm_output.shape[1:]
            derivative_size = np.prod(tlm_shape[output_shape_indices:])

            Telem = MixedElement([self.output_space.ufl_element()] * derivative_size)
            T = FunctionSpace(self.output_space.mesh(), Telem)

            t = Function(T)
            t.vector()[:] = tlm_output.flatten()
            t.tlm_shape = tlm_shape

            rv[j] = t
        return rv

    def recompute(self):
        deps = self.get_dependencies()
        net = deps[0].saved_output
        args = [bv.saved_output for bv in deps[1:]]
        output = backend_run_network(net, self.output_space, *args, **self.kwargs)
        self.get_outputs()[0].checkpoint = output



from pyadjoint.enlisting import Enlist
from pyadjoint.tape import get_working_tape, stop_annotating
import numpy
from pyadjoint.overloaded_type import OverloadedType
from typing import Union, List, Optional, Any
from .adjfloat import AdjFloat
from .identity import JacobianIdentity, make_jacobian_identities
from .jax_adjoint import ndarray
from .control import Control
from .tape import Tape

Array = Any


def compute_gradient(
    J: Union[List[OverloadedType], OverloadedType],
    m: Union[List[Control], Control],
    options: Optional[dict] = None,
    tape: Optional[Tape] = None,
    adj_value: float = 1.0,
) -> Union[List[OverloadedType], OverloadedType]:
    """
    Compute the gradient of J with respect to the initialisation value of m,
    that is the value of m at its creation.

    Args:
        J (OverloadedType, list[OverloadedType]):  The objective functional.
        m (Union[list[Control], Control]): The (list of) controls.
        options (dict): A dictionary of options. To find a list of available options
            have a look at the specific control type.
        tape (Tape): The tape to use. Default is the current tape.

    Returns:
        OverloadedType: The derivative with respect to the control. Should be an instance of the same type as
            the control.
    """
    options = {} if options is None else options
    tape = get_working_tape() if tape is None else tape
    tape.reset_variables()
    adj_value = Enlist(adj_value)
    J = Enlist(J)
    m = Enlist(m)

    for i in range(len(adj_value)):
        J[i].block_variable.adj_value = adj_value[i]

    with stop_annotating():
        with tape.marked_nodes(m):
            tape.evaluate_adj(markings=True, inputs=m, outputs=J)

    grads = [i.get_derivative(options=options) for i in m]
    return m.delist(grads)


def compute_jacobian_matrix(
    J: Union[List[OverloadedType], OverloadedType],
    m: Union[List[Control], Control],
    m_jac: Any = None,
    tape: Tape = None,
) -> Any:
    """
    Compute dJdm matrix.

    Args:
        J (OverloadedType): The outputs of the function.
        m (list[pyadjoint_utils.Control] or pyadjoint_utils.Control): The (list of) controls.
        m_jac: An input Jacobian to multiply with. By default, this will be an identity Jacobian.
            If m is a list, this should be a list of lists with len(m_jac) == len(m) and
            len(m_jac[i]) == len(m) for each i-th entry in m_jac.
        tape: The tape to use. Default is the current tape.

    Returns:
        OverloadedType: The jacobian with respect to the control. Should be an instance of the same type as the control.
    """
    tape = get_working_tape() if tape is None else tape

    m = Enlist(m)
    if m_jac is None:
        m_jac = make_jacobian_identities(len(m))
    else:
        m_jac = Enlist(m_jac)

    tape.reset_tlm_matrix_values()

    J = Enlist(J)

    for i, input_jac in enumerate(m_jac):
        m[i].tlm_matrix = Enlist(input_jac)

    with stop_annotating():
        tape.evaluate_tlm_matrix(m, J)

    r = [v.block_variable.tlm_matrix for v in J]
    return J.delist(r)


def compute_jacobian_action(
    J: Union[List[OverloadedType], OverloadedType],
    m: Union[List[Control], Control],
    m_dot: Union[List[OverloadedType], OverloadedType],
    options: Optional[dict] = None,
    tape: Optional[Tape] = None,
) -> Union[List[OverloadedType], OverloadedType]:
    """
    Compute the action of the Jacobian of J on m_dot with respect to the
    initialisation value of m, that is the value of m at its creation.

    Args:
        J (OverloadedType):  The outputs of the function.
        m (list[pyadjoint.Control] or pyadjoint.Control): The (list of) controls.
        options (dict): A dictionary of options. To find a list of available options
            have a look at the specific control type.
        tape: The tape to use. Default is the current tape.
        m_dot(OverloadedType): variation of same overloaded type as m.

    Returns:
        OverloadedType: The action on m_dot of the Jacobian of J with respect to the control. Should be an instance of the same type as the output of J.
    """
    options = {} if options is None else options
    tape = get_working_tape() if tape is None else tape
    tape.reset_tlm_values()
    m_dot = Enlist(m_dot)
    J = Enlist(J)
    m = Enlist(m)

    for i in range(len(m_dot)):
        m[i].tlm_value = m_dot[i]

    with stop_annotating():
        tape.evaluate_tlm(inputs=m, outputs=J)

        Jmdots = []
        for Ji in J:
            if isinstance(Ji.block_variable.tlm_value, numpy.ndarray):
                output = Ji.block_variable.output._ad_copy()
                output, offset = output._ad_assign_numpy(
                    output, Ji.block_variable.tlm_value.flatten(), offset=0
                )
            else:
                output = Ji.block_variable.tlm_value
            Jmdots.append(output)
    return J.delist(Jmdots)


def compute_hessian_action(
    J: Union[List[OverloadedType], OverloadedType],
    m: Union[List[Control], Control],
    m_dot: Union[List[OverloadedType], OverloadedType],
    options: Optional[dict] = None,
    tape: Optional[Tape] = None,
    hessian_value=1.0,
) -> Union[List[OverloadedType], OverloadedType]:
    """
    Compute the Hessian of J in a direction m_dot at the current value of m

    Args:
        J (AdjFloat):  The objective functional.
        m (list or instance of Control): The (list of) controls.
        m_dot (list or instance of the control type): The direction in which to compute the Hessian.
        options (dict): A dictionary of options. To find a list of available options
            have a look at the specific control type.
        tape: The tape to use. Default is the current tape.

    Returns:
        OverloadedType: The second derivative with respect to the control in direction m_dot. Should be an instance of
            the same type as the control.
    """
    tape = get_working_tape() if tape is None else tape
    options = {} if options is None else options

    tape.reset_tlm_values()
    tape.reset_hessian_values()

    m = Enlist(m)
    m_dot = Enlist(m_dot)
    for i, value in enumerate(m_dot):
        m[i].tlm_value = m_dot[i]

    with stop_annotating():
        tape.evaluate_tlm()

    hessian_value = Enlist(hessian_value)
    J = Enlist(J)
    for i in range(len(hessian_value)):
        J[i].block_variable.hessian_value = hessian_value[i]

    with stop_annotating():
        with tape.marked_nodes(m):
            tape.evaluate_hessian(markings=True)

    r = [v.get_hessian(options=options) for v in m]
    return m.delist(r)



from .equation_solver import ReducedEquationSolver

from pyadjoint.tape import (
    get_working_tape,
    set_working_tape,
    stop_annotating,
    no_annotations,
    annotate_tape,
)
from pyadjoint.block import Block
from pyadjoint.overloaded_type import create_overloaded_object, OverloadedType
from pyadjoint.enlisting import Enlist
from pyadjoint_utils import *

from petsc4py import PETSc
import numpy as np


def _get_enum_lookup(enum, lookup=None):
    lookup = lookup if lookup is not None else {}
    names = (name for name in dir(enum) if not name.startswith("_"))
    for name in names:
        val = getattr(enum, name)
        lookup[val] = name
    return lookup


# Build lookup tables for printing out the converged reasons.
_defaults = {
    -3: "DIVERGED_LINEAR_SOLVE",
}
snes_converged_reason_lookup = _get_enum_lookup(PETSc.SNES.ConvergedReason, _defaults)


_defaults = {
    -11: "DIVERGED_PC_FAILED",
    -3: "DIVERGED_ITS",
}
ksp_converged_reason_lookup = _get_enum_lookup(PETSc.KSP.ConvergedReason, _defaults)


class SNESSolver(ReducedEquationSolver):
    def __init__(self, eq, parameters=None):
        ReducedEquationSolver.__init__(self, eq, parameters)

        self.rf_np = ReducedFunctionNumPy(self.eq.reduced_function)

        param_copy = self.parameters.copy()

        self.jmat_type = param_copy.pop("jmat_type", "assembled")
        self.pmat_type = param_copy.pop("pmat_type", "J")
        self.adj_jmat_type = param_copy.pop("adj_jmat_type", self.jmat_type)
        self.adj_pmat_type = param_copy.pop("adj_pmat_type", "J")
        self.assembled_force_dense = param_copy.pop("assembled_force_dense", True)

        if self.jmat_type not in ("action", "assembled"):
            raise ValueError("Invalid jmat_type: %s" % self.jmat_type)
        if self.pmat_type not in ("J", "assembled"):
            raise ValueError("Invalid pmat_type: %s" % self.pmat_type)

        if self.adj_jmat_type not in ("action", "assembled"):
            raise ValueError("Invalid adj_jmat_type: %s" % self.adj_jmat_type)
        if self.adj_pmat_type not in ("J", "assembled"):
            raise ValueError("Invalid adj_pmat_type: %s" % self.adj_pmat_type)

        if len(param_copy) > 0:
            import warnings

            warnings.warn(
                "Extra parameters given to SNESSolver: " + ", ".join(parameters.keys()),
                stacklevel=2,
            )

    def solve(self, solution_controls, **kwargs):
        annotate = annotate_tape(kwargs)
        solution_controls = Enlist(solution_controls)

        if annotate:
            tape = get_working_tape()
            if tape == self.eq.reduced_function.tape:
                raise ValueError("reduced_function must be recorded on a separate tape")

            b_kwargs = SNESSolveBlock.pop_kwargs(kwargs)
            b_kwargs.update(kwargs)
            block = SNESSolveBlock(self, solution_controls, **b_kwargs)

        # Compute solution.
        solution = self.backend_solve(solution_controls, **kwargs)

        for i, obj in enumerate(solution):
            solution[i] = create_overloaded_object(obj)

        if annotate:
            for obj in solution:
                block.add_output(obj.create_block_variable())
            tape.add_block(block)

        return solution_controls.delist(solution)

    @no_annotations
    def backend_solve(self, solution_controls, **kwargs):
        self.snes_solve(solution_controls, **kwargs)
        return [p.data() for p in self.eq.reduced_function.controls]

    class SNESInterface(object):
        def __init__(self, rf_np, bc, h_bc, dense=True, disp=True, cback=None):
            self.rf_np = rf_np
            self.bc = bc
            self.h_bc = h_bc
            self.dense = dense
            self.disp = disp
            self.cback = cback
            self.i = 0
            self.mults = 0

        def formFunction(self, snes, X, F):
            # Apply boundary conditions.
            X_np = X[:]

            X_rf = self.bc(X_np, return_numpy=False)

            if self.cback is not None:
                self.cback(X_rf)

            X_bc_np = self.rf_np.get_outputs_array(Enlist(X_rf))
            res_np = X_np - X_bc_np

            # Evaluate residual.
            res_rf = self.rf_np.rf(X_rf)

            # Apply homogeneous boundary conditions.
            res_np += self.h_bc(res_rf, adjoint=True, input_numpy=False)

            F[:] = res_np
            F.assemble()
            if self.disp:
                print("%2d: ||residual||: % 18.15g" % (self.i, np.linalg.norm(res_np)))
            self.i += 1

        def formJacobian(self, snes, X, J, P):
            P.zeroEntries()
            if self.disp:
                print("%2d: Jacobian" % (self.i))
            res_np = self.rf_np(X[:])
            J_dolfin = self.rf_np.jac_matrix()
            J_dolfin = self.bc(J_dolfin, False, False)
            if self.disp:
                J_np = J_dolfin.array()
                s = np.linalg.svd(J_np, full_matrices=False, compute_uv=False)
                print("%2d: svd: " % (self.i), s)

            if self.dense:
                J[:, :] = J_dolfin.array()
            else:
                J_petsc = J_dolfin.instance().mat()
                J_petsc.copy(J)
            J.assemble()

        def formPreconditioner(self, snes, X, J, P):
            if self.disp:
                print("%2d: Preconditioner" % (self.i), J, P)
            res_np = self.rf_np(X[:])
            J_dolfin = self.rf_np.jac_matrix()
            J_dolfin = self.bc(J_dolfin, False, False)
            if self.disp:
                J_np = J_dolfin.array()
                s = np.linalg.svd(J_np, full_matrices=False, compute_uv=False)
                print("%2d: svd: " % (self.i), s)

            if self.dense:
                P[:, :] = J_dolfin.array()
            else:
                J_petsc = J_dolfin.instance().mat()
                J_petsc.copy(P)
            P.assemble()

        def dummyJacobian(self, snes, X, J, P):
            """Uses X to set the values on the tape so that the correct Jacobian is generated."""
            if self.disp:
                print("Finished Jacobian mults: %d/%d" % (self.mults, X.size))
                print("%2d: dummy Jacobian" % self.i)
            m = self.rf_np(X[:])
            self.mults = 0

        def mult(self, mat, dx_vec, y):
            self.mults += 1

            # Apply homogeneous boundary conditions.
            dx_rf = self.h_bc(dx_vec[:], return_numpy=False)

            y_rf = self.rf_np.rf.jac_action(dx_rf)

            # Apply homogeneous boundary conditions.
            y_np = self.h_bc(y_rf, input_numpy=False)

            y.setArray(y_np)

        def multTranspose(self, mat, dx_vec, y):
            self.mults += 1

            # Apply homogeneous boundary conditions.
            dx_rf = self.h_bc(dx_vec[:], return_numpy=False)

            y_rf = self.rf_np.rf.adj_jac_action(dx_rf)

            # Apply homogeneous boundary conditions.
            y_np = self.h_bc(y_rf, input_numpy=False)

            y.setArray(y_np)

    def bc(self, x, input_numpy=True, return_numpy=True):
        if input_numpy:
            x_rf = self.rf_np.get_rf_input(x)
            x_rf = self.rf_np.controls.delist(x_rf)
        else:
            x_rf = x

        for bc in self.eq.bcs:
            x_rf = bc(x_rf)

        if return_numpy:
            x_np = self.rf_np.get_outputs_array(Enlist(x_rf))
            return x_np
        return x_rf

    def h_bc(self, x, adjoint=False, input_numpy=True, return_numpy=True):
        if input_numpy:
            if adjoint:
                x_rf = self.rf_np.get_rf_adj_input(x)
                x_rf = self.rf_np.outputs.delist(x_rf)
            else:
                x_rf = self.rf_np.get_rf_input(x)
                x_rf = self.rf_np.controls.delist(x_rf)
        else:
            x_rf = x

        # This assumes that the h_bc works on both the rf's inputs and its outputs.
        for bc in self.eq.h_bcs:
            x_rf = bc(x_rf)

        if return_numpy:
            if adjoint:
                x_np = self.rf_np.get_global(x_rf)
            else:
                x_np = self.rf_np.get_outputs_array(x_rf)
            return x_np
        return x_rf

    def setup_snes(
        self,
        m_global,
        disp=False,
        ksp_type="gmres",
        pc_type="none",
        rtol=None,
        atol=None,
        stol=None,
        max_it=None,
        cback=None,
    ):

        snes_interface = self.SNESInterface(
            self.rf_np,
            self.bc,
            self.h_bc,
            dense=self.assembled_force_dense,
            disp=disp,
            cback=cback,
        )

        n = m_global.size

        snes = PETSc.SNES().create()

        if self.jmat_type == "assembled":
            J_dolfin = self.rf_np.jac_matrix()
            if self.assembled_force_dense:
                J = PETSc.Mat()
                J.createDense(n)
                J.setUp()
            else:
                J = J_dolfin.instance().mat()
                J_dolfin = self.bc(J_dolfin, False, False)
            snes.setJacobian(snes_interface.formJacobian, J)
        elif self.jmat_type == "action":
            J = PETSc.Mat()
            J.createPython(n, snes_interface)
            J.setUp()
            if self.pmat_type == "assembled":
                P = PETSc.Mat()
                P.createDense(n)
                P.setUp()
                raise ValueError(
                    'pmat_type "assembled" cannot be used with jmat_type "action" yet'
                )
                # snes.setJacobian(snes_interface.formPreconditioner, J=J, P=P)
            elif self.pmat_type == "J":
                P = J
            else:
                raise ValueError("Invalid pmat_type: %s" % self.pmat_type)
            snes.setJacobian(snes_interface.dummyJacobian, J, P)
        else:
            raise ValueError("Invalid jmat_type: %s" % self.jmat_type)

        X, F = J.createVecs()
        snes.setFunction(snes_interface.formFunction, F)
        snes.setTolerances(rtol=rtol, atol=atol, stol=stol, max_it=max_it)
        self._tolerances = {"rtol": rtol, "atol": atol, "stol": stol, "max_it": max_it}
        self._ksp_type = ksp_type
        ksp = snes.getKSP()
        ksp.setType(ksp_type)
        self._pc_type = pc_type
        pc = ksp.getPC()
        pc.setType(pc_type)
        snes.setFromOptions()

        return snes, X, F, J

    @no_annotations
    def snes_solve(self, solution_controls, **kwargs):
        disp = kwargs.get("disp", False)

        m_global = self.rf_np.get_controls(solution_controls)
        # m_global = self.rf_np.obj_to_array([p.tape_value() for p in solution_controls])

        snes, X, F, J = self.setup_snes(m_global, **kwargs)

        X[:] = self.bc(m_global)

        snes.solve(None, X)

        reason = snes.getConvergedReason()
        if reason < 0:
            # Hack to get an optimization line search to maybe keep going.
            X[:] = np.full_like(X[:], np.inf)

        reasonMessage = snes_converged_reason_lookup.get(
            reason, "UNKNOWN (%s)" % reason
        )

        if disp or reason < 0:
            if reason < 0:
                print("SNES solver failed with reason %s" % reasonMessage)
            else:
                print("SNES Converged reason:", reasonMessage)
            print("SNES Num iterations:", snes.getIterationNumber())
            if reason < 0 and reasonMessage == "DIVERGED_LINEAR_SOLVE":
                ksp = snes.getKSP()
                reason = ksp.getConvergedReason()
                reasonMessage = ksp_converged_reason_lookup.get(
                    reason, "UNKNOWN (%s)" % reason
                )

                if disp or reason < 0:
                    if reason < 0:
                        print("  KSP solver failed with reason %s" % reasonMessage)
                    else:
                        print("  KSP Converged reason:", reasonMessage)
                    print("  KSP Num iterations:", ksp.getIterationNumber())
            print()
        self.rf_np.set_controls(self.bc(X[:]))

    class AdjJContext(object):
        def __init__(self, rf_np, bc):
            self.rf_np = rf_np
            self.bc = bc

        def mult(self, mat, dx_vec, y):
            dx_rf = self.bc(dx_vec[:], return_numpy=False)
            y_rf = self.rf_np.rf.adj_jac_action(dx_rf)
            y_np = self.bc(y_rf, input_numpy=False)
            y.setArray(y_np)

    @no_annotations
    def _adjoint_ksp_solve(self, adj_inputs, disp=False):
        m_global = self.rf_np.get_outputs_array(adj_inputs)
        n = m_global.size

        # Set up solver.
        ksp = PETSc.KSP().create()
        ksp.setOptionsPrefix("adj_")

        pc = ksp.getPC()
        ksp.setType(self._ksp_type)
        pc.setType(self._pc_type)
        ksp.setFromOptions()

        if self.adj_jmat_type == "assembled":
            J_dolfin = self.rf_np.jac_matrix()

            # Transpose petsc4py Mat local-to-global mapping.
            # Note: this won't work for complex values and maybe not for multiple processes.
            J = J_dolfin.instance().mat()
            lgmap = J.getLGMap()
            J.transpose()
            J.setLGMap(lgmap[1], lgmap[0])
            J_dolfin = self.bc(J_dolfin, False, False)

            if self.assembled_force_dense:
                if disp:
                    J_np = J_dolfin.array()
                    s = np.linalg.svd(J_np, full_matrices=False, compute_uv=False)
                    print("adj: svd: ", s)
                J = PETSc.Mat()
                J.createDense(n)
                J.setUp()
                J[:, :] = J_dolfin.array().conj()
                J.assemble()

            if disp:
                J_np = J_dolfin.array()
                s = np.linalg.svd(J_np, full_matrices=False, compute_uv=False)
                print("adj: svd: ", s)
            ksp.setOperators(J)
        elif self.adj_jmat_type == "action":
            ksp_interface = self.AdjJContext(self.rf_np, self.h_bc)
            J = PETSc.Mat()
            J.createPython(n, ksp_interface)
            J.setUp()
            if self.adj_pmat_type == "assembled":
                P = PETSc.Mat()
                P.createDense(n)
                P.setUp()
                raise ValueError(
                    'adj_pmat_type "assembled" cannot be used with adj_jmat_type "action" yet'
                )
            elif self.adj_pmat_type == "J":
                P = J
            else:
                raise ValueError("Invalid adj_pmat_type: %s" % self.adj_pmat_type)
            ksp.setOperators(J, P)
        else:
            raise ValueError("Invalid adj_jmat_type: %s" % self.adj_jmat_type)

        # Set up the system's vectors.
        X, F = J.createVecs()
        X.set(0)
        F[:] = self.h_bc(m_global[:])

        # Solve.
        ksp.solve(F, X)
        reason = ksp.getConvergedReason()
        reasonMessage = ksp_converged_reason_lookup.get(reason, "UNKNOWN (%s)" % reason)

        if disp or reason < 0:
            if reason < 0:
                print("KSP solver failed with reason %s" % reasonMessage)
            else:
                print("KSP Converged reason:", reasonMessage)
            print("KSP Num iterations:", ksp.getIterationNumber())
            print()

        # Convert X to list of controls
        adj_sol = self.rf_np.get_rf_input(X[:])
        return adj_sol


class SNESSolveBlock(Block):

    pop_kwargs_keys = ["adj_cb"]

    def __init__(self, solver, solution_controls, **kwargs):
        super(SNESSolveBlock, self).__init__()
        self.adj_cb = kwargs.pop("adj_cb", None)

        self.solver = solver
        self.rf = self.solver.eq.reduced_function
        self.solution_controls = solution_controls
        self.forward_kwargs = kwargs

        # The dependencies consist of every block variable that contributes to the rf's output but
        # is not dependendent on the initial guess.
        with self.rf.tape.marked_nodes(self.solution_controls):
            dependencies, blocks = self.rf.tape.find_relevant_dependencies(
                self.rf.outputs
            )
            for dep in dependencies:
                if not dep.marked_in_path:
                    self.add_dependency(dep.output)

        # It might be nice to recompute the solution based on the initial guess, too.
        self.solution_block_variables = []
        for c in self.solution_controls:
            self.add_dependency(c.control)
            self.solution_block_variables.append(c.block_variable)

    def __str__(self):
        return "SNESSolveBlock"

    def reset_variables(self, types=None):
        super().reset_variables(types)
        self.rf.tape.reset_variables(types)

    def prepare_recompute_component(self, inputs, relevant_outputs):
        solution = self.solver.backend_solve(
            self.solution_controls, **self.forward_kwargs
        )
        return solution

    def recompute_component(self, inputs, block_variable, idx, prepared):
        solution = prepared
        return solution[idx]

    def prepare_evaluate_adj(self, inputs, adj_inputs, relevant_dependencies):
        # Important: the rf tape needs to be reset because the _adjoint_ksp_solve needs to use the adj_values.
        # But we don't want to lose the adjoint values that have already been computed on the working tape.
        # So this saves the current adjoint values before doing the adjoint solve and then reinstates those values.
        tape = self.rf.tape
        with tape.save_adj_values():
            adj_sol = self.solver._adjoint_ksp_solve(adj_inputs)

        if self.adj_cb is not None:
            self.adj_cb(adj_sol)

        relevant_block_variables = [bv for i, bv in relevant_dependencies]

        # Pass the adj_sol to the rf tape to calculate the adjoint of all the controls.
        # FIXME: A problem here is that not all block variables use the same
        #        type in their adj_value as in their normal overloaded type.
        #        E.g., Function types use dolfin vectors to pass the adj_value.
        with tape.save_adj_values():
            for i, bv in enumerate(self.rf.outputs):
                bv.adj_value = create_overloaded_object(adj_sol[i])._ad_mul(-1)

            with stop_annotating():
                with tape.marked_nodes(relevant_block_variables):
                    tape.evaluate_adj(markings=True)

            # Extract adjoint values from controls.
            adj_values = [bv.adj_value for bv in self.get_dependencies()]
        return adj_values

    def evaluate_adj_component(
        self, inputs, adj_inputs, block_variable, idx, prepared=None
    ):
        return prepared[idx]

    def tf_add_extra_to_graph(self, tf_tensors):
        import tensorflow as tf

        self.rf.tape._tf_tensors.update(tf_tensors)
        # Add rf tape as a sub-block.
        with tf.name_scope("Residual_tape"):
            self.rf.tape._tf_add_blocks()

        tf_tensors.update(self.rf.tape._tf_tensors)



from crikit.cr.types import Space


class DirectSum(Space):
    """This space represents a concatenation of separate spaces."""

    def __init__(self, *spaces):
        if len(spaces) == 1 and isinstance(spaces[0], (tuple, list)):
            spaces = tuple(spaces[0])
        self._spaces = spaces
        if all((space == self._spaces[0]) for space in self._spaces):
            self._shape = (len(self._spaces),) + self._spaces[0].shape()
        else:
            self._shape = tuple(space.shape() for space in self._spaces)
        self._enlisted = False

    def __repr__(self):
        return "DirectSum" + str(self._spaces)

    def __eq__(self, other):
        return (
            isinstance(other, DirectSum)
            and self._shape == other._shape
            and all(mine == theirs for mine, theirs in zip(self._spaces, other._spaces))
        )

    def delist(self, y=None):
        y = self if y is None else y
        if self._enlisted:
            return y[0]
        else:
            return y

    def __getitem__(self, idx):
        """Can be used to iterate through the spaces."""
        return self._spaces[idx]

    def __len__(self):
        """Returns the number of component spaces"""
        return len(self._spaces)

    def point(self, **kwargs):
        subargs = kwargs.get("subargs", tuple(() for space in self._spaces))
        subkwargs = kwargs.get("subkwargs", tuple(dict() for space in self._spaces))
        if "near" in kwargs:
            near_point = kwargs["near"]
            for p, kw in zip(near_point, subkwargs):
                kw["near"] = p
        if "zero" in kwargs:
            zero = kwargs["zero"]
            for kw in subkwargs:
                kw["zero"] = zero
        return tuple(
            space.point(*sargs, **skwargs)
            for (space, sargs, skwargs) in zip(self._spaces, subargs, subkwargs)
        )

    def is_point(self, tup):
        if not isinstance(tup, (list, tuple)) or len(tup) != len(self._spaces):
            return False
        return all(space.is_point(point) for space, point in zip(self._spaces, tup))

    def shape(self):
        return self._shape


def enlist(spaces):
    """Given a space or a tuple of spaces, this returns a DirectSum representing
    those spaces. If the given space is already a DirectSum, it returns that
    space. Otherwise, it creates a DirectSum of the space(s) and marks it as
    enlisted so that it can delisted later.

    Args:
        spaces (Space, tuple[Space], or list[Space]): the space(s) to enlist

    Returns:
        DirectSum: a single space containing the given space(s)
    """
    if isinstance(spaces, DirectSum):
        return spaces
    else:
        dspace = DirectSum(spaces)
        dspace._enlisted = True
        return dspace


class Multiset(Space):
    """This space represents a single space repeated multiple times."""

    def __init__(self, space, n):
        self._space = space
        self._n = n
        self._shape = (n,) + space.shape()

    @property
    def space(self):
        """the base space for the Multiset"""
        return self._space

    def __repr__(self):
        return "Multiset(" + repr(self._space) + "," + repr(self._n) + ")"

    def __getitem__(self, idx):
        """Can be used to iterate through the spaces."""
        if 0 <= idx and idx < self._n:
            return self._space
        raise IndexError("index out of range")

    def __eq__(self, other):
        if (
            isinstance(other, Multiset)
            and self._space == other._space
            and self._shape == other._shape
        ):
            return True
        return False

    def __len__(self):
        """Returns the number of component spaces"""
        return self._n

    def is_point(self, tup):
        if not isinstance(tup, tuple) or len(tup) != len(self._spaces):
            return False
        return all(self._space.is_point(point) for point in tup)

    def shape(self):
        return self._shape



from jax import jit
import jax.numpy as np
import numpy as onp
from jax.tree_util import Partial as partial
from itertools import chain, combinations


@jit
def symm(x):
    """Symmetrizes the input

    :param x: a 2-d array to symmetrize
    :type x: Union[np.ndarray,onp.ndarray]
    :return: A symmetric (and doubled) version of ``x``
    :rtype: Union[np.ndarray,onp.ndarray]

    """
    return x + x.T


@jit
def antisymm(x):
    """Antisymmetrizes the input

    :param x: a 2-d array to antisymmetrize
    :type x: Union[np.ndarray,onp.ndarray]
    :return: An antisymmetric (and doubled) version of ``x``
    :rtype: Union[np.ndarray,onp.ndarray]

    """
    return x - x.T


@jit
def commutator_action(A, B, v):
    bv = B @ v
    av = A @ v
    return A @ bv - B @ av


@jit
def anticommutator_action(A, B, v):
    Av = A @ v
    Bv = B @ v
    return A @ Bv + B @ Av


@jit
def scalar_triple_prod(u, v, w):
    return np.dot(v, np.cross(u, w))


# Copied from https://docs.python.org/3/library/itertools.html
def powerset(iterable, exclude_empty_set=True):
    "powerset([1,2,3], False) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
    s = tuple(iterable)
    return chain.from_iterable(
        combinations(s, r) for r in range(int(exclude_empty_set), len(s) + 1)
    )


def levi_civita(n):
    """Returns the Levi-Civita pseudotensor in ``n`` dimensions.

    :param n: the number of dimensions
    :type n: int
    :returns: The Levi-Civita pseudotensor in ``n`` spatial dimensions
    :rtype: np.ndarray

    """
    if n == 2:
        return np.array([[0, 1], [-1, 0]])
    elif n == 3:
        eps = onp.zeros((3, 3, 3))
        pos_idx = [(0, 1, 2), (1, 2, 0), (2, 0, 1)]
        neg_idx = [(2, 1, 0), (0, 2, 1), (1, 0, 2)]
        for ix, iy, iz in pos_idx:
            eps[ix, iy, iz] = 1

        for ix, iy, iz in neg_idx:
            eps[ix, iy, iz] = -1

        return np.array(eps)
    else:
        raise NotImplementedError


eps_ijk = levi_civita(3)
eps_ij = levi_civita(2)


@jit
def _eps_vec_action(v):
    return np.einsum("ijk,k -> ij", eps_ijk, v)


@jit
def axial_vector(G):
    return np.einsum("ijk,jk->i", eps_ijk, G, optimize=True)


@jit
def _tprod(x, y):
    return np.tensordot(x, y, axes=0)


def symm_q4(T):
    S = np.zeros_like(T)
    for i in range(3):
        for j in range(3):
            for k in range(3):
                for l in range(3):
                    idx = (i, j, k, l)
                    for pi in permutations(idx):
                        S = jax.ops.index_add(S, idx, T[tuple(pi)])

    return S


@jit
def symm_q3(T):
    return (
        T + np.moveaxis(T, [0, 1, 2], [1, 2, 0]) + np.moveaxis(T, [0, 1, 2], [2, 0, 1])
    )


@jit
def tA(T, A):
    return np.einsum("ijk,jk -> i", T, A, optimize=True)


@jit
def tv(T, v):
    return tA(T, _tprod(v, v))


@jit
def Tv(T, v):
    return np.einsum("ijk,k -> ij", T, v, optimize=True)


@jit
def TinnerS(T, S):
    return np.einsum("ijk,jkl -> il", T, S, optimize=True)


@jit
def TcontrS(T, S):
    return np.einsum("ijk,ijk", T, S, optimize=True)


@jit
def TW(T, W):
    return np.einsum("ijm,mk -> ijk", T, W, optimize=True)


@jit
def tbrace(T):
    # eq 2.34 in Zheng '94
    return [
        T,
        np.moveaxis(T, [1, 2, 3], [2, 3, 1]),
        np.moveaxis(T, [1, 2, 3], [3, 1, 2]),
    ]


@jit
def near(y, x, tol):
    return np.abs(x - y) <= tol


@jit
def _is_third_order_irreducible(T):
    if len(T.shape) != 3:
        return False

    t1 = T[1, 1, 1]
    t2 = T[2, 2, 2]
    t1id = (1, 1, 1)
    t2id = (2, 2, 2)
    t1idx = {(1, 2, 2), (2, 1, 2), (2, 2, 1)}
    t2idx = {(2, 1, 1), (1, 2, 1), (1, 1, 2)}
    for i in range(t.shape[0]):
        for j in range(t.shape[1]):
            for k in range(t.shape[2]):
                idx = (i, j, k)
                if idx == t1id or idx == t2id:
                    continue

                if idx in t1idx:
                    if not near(T[idx], -t1, 1.0e-10):
                        return False
                elif idx in t2idx:
                    if not near(T[idx], -t2, 1.0e-10):
                        return False
                else:
                    if not near(T[idx], 0.0, 1.0e-10):
                        return False

    return True


@partial(jit, static_argnums=(1,))
def spectral_decomp(A, hermitian=False):
    w, V = np.linalg.eigh(A) if hermitian else np.linalg.eig(A)
    return V, np.diag(w)


def factorial(n: int):
    if n == 0:
        return 1
    return np.prod(np.arange(1, n + 1))


# @partial(jit,static_argnums=(2,))
def matpows(A, ns, hermitian=False):
    # TODO: use a better algorithm for this, maybe exponentiation by squaring?
    V, D = spectral_decomp(A, hermitian)
    Vinv = np.linalg.inv(V)
    return [Vinv @ np.power(D, i) @ V for i in ns]


def matexp(A, hermitian=False):
    # TODO: find actually good choices of these parameters
    N = range(10)
    return sum([A_i / factorial(i) for A_i, i in zip(matpows(A, N, hermitian), N)])


@jit
def _3d_rotation_matrix(axis, theta):
    ct = np.cos(theta)
    st = np.sin(theta)
    ux = axis[0]
    uy = axis[1]
    uz = axis[2]
    return np.array(
        [
            [
                ct + ux * ux * (1 - ct),
                ux * uy * (1 - ct) - uz * st,
                ux * uz * (1 - ct) + uy * st,
            ],
            [
                uy * ux * (1 - ct) + uz * st,
                ct + uy * uy * (1 - ct),
                uy * uz * (1 - ct) - ux * st,
            ],
            [
                uz * ux * (1 - ct) - uy * st,
                uz * uy * (1 - ct) + ux * st,
                ct + uz * uz * (1 - ct),
            ],
        ]
    )


@jit
def householder_matrix(v):
    return np.eye(v.size) - _tprod(v, v)


def near(val, to, rtol=1.0e-5):
    """Returns True if ``val`` and ``to`` are within relative tolerance ``rtol``
    and False otherwise

    :param val: a value
    :type val: np.ndarray
    :param to: is ``val`` close to this?
    :type to: np.ndarray
    :param rtol: Relative tolerance, defaults to 1.0e-5
    :type rtol: float, optional
    :return: are ``val`` and ``to`` within ``rtol``?
    :rtype: bool

    """
    # effectively a JAX DeviceArray( ,dtype=bool) to bool conversion
    return bool(np.allclose(val, to, rtol=rtol))


def is_symm(X, rtol=1.0e-5):
    return near(X, 0.5 * symm(X), rtol)


def is_antisymm(X, rtol=1.0e-5):
    return near(X, 0.5 * antisymm(X), rtol)



from crikit.cr.types import PointMap
from crikit.cr.space_builders import DirectSum, Multiset, enlist
from pyadjoint.enlisting import Enlist
from functools import update_wrapper
from itertools import chain
from typing import Union, Iterable, TypeVar, Optional, List

# these TypeVars represent generic points in the input, output,
# and parameter spaces of various PointMaps
InputPoint = TypeVar("InputPoint")
ParameterPoint = TypeVar("ParameterPoint")
OutputPoint = TypeVar("OutputPoint")


class Callable(PointMap):
    """This class wraps a Python callable into a PointMap.

    I.e., if you (1) have a function or a class/object with a ``__call__`` method,
    and you (2) know the input space and output space of the callable, then this
    will put the callable into the Space/PointMap structure.

    Set bare to true if the callable requires that the point be unpacked.

    A point map that accepts multiple inputs will receive those inputs as a
    single tuple, so if your callable expects to receive them as separate
    arguments, you must use ``bare=True``.

        >>> from crikit.cr.map_builders import Callable
        >>> source = ... # The source and target space don't really matter here.
        >>> target = ...
        >>> standard_callable = lambda x: x[0] + x[1] + x[2]
        >>> c = Callable(source, target, standard_callable)
        >>> c((1, 2, 3))
        6
        >>> bare_callable = lambda a, b, c: a + b + c
        >>> c_bare = Callable(source, target, bare_callable, bare=True)
        >>> c_bare((1, 2, 3))
        6
        >>> c_bad = Callable(source, target, bare_callable, bare=False)
        >>> c_bad((1, 2, 3))
        Traceback (most recent call last):
        ...
        TypeError: <lambda>() missing 2 required positional arguments: 'b' and 'c'
        >>> c_bad(1, 2, 3)
        Traceback (most recent call last):
        ...
        TypeError: __call__() takes 2 positional arguments but 4 were given

    TODO:
        * Define a function decorator to make it easier to use.
    """

    def __init__(self, source_space, target_space, callble, bare=False):
        assert callable(callble)
        super(Callable, self).__init__(source_space, target_space)
        self._callable = callble
        self._bare = bare
        update_wrapper(self, self._callable)

    @property
    def callable(self):
        """the callable that this point map was initialized with"""
        return self._callable

    def __call__(self, point, **kwargs):
        """Calls the callable with given point, unpacking the point first if
        initialized with ``bare=True``"""
        if self._bare:
            return self._callable(*point, **kwargs)
        else:
            return self._callable(point, **kwargs)


class AugmentPointMap(PointMap):
    """This class wraps a PointMap so that keyword arguments of its ``__call__``
    method are moved to the explicit input space of the map.

    In mathematical notation, this wraps a function :math:`f(u; p)` so that it must be called as :math:`f(u, p)`.

    If the original point map is called as :code:`point_map(point)`, then the augmented point map
    is called as :code:`aug_map((point, params))` (or :code:`aug_map((point, *params)` if :code:`bare=True`).

    Args:
        point_map (PointMap): The point map to wrap.
        param_names (str, tuple[str], or list[str]): The keywords that will be
            added to the input space of the point map.
        param_space (Space, tuple[Space], or list[Space]): The spaces
            corresponding to each keyword in the param_names list.
        bare (bool): If you set this to true, then the ``__call__`` method
            expects the params to be unpacked.

    TODO:
        * Rename the class to ``Augment`` to match Callable and Parametric?
            * Or rename the other ones to match CompositePointMap and ParallelPointMap?
        * Maybe rename bare to bare_params and add a new parameter called bare_point.
          If they're both true, then map can be called as ``aug_map((*point, *params))``.
    """

    def __init__(self, point_map, param_names, param_space, bare=False):
        param_names = Enlist(param_names)
        if bare:
            param_space = enlist(param_space)
            self._param_len = len(param_space)
            new_source = DirectSum(tuple(chain(enlist(point_map.source), param_space)))
        else:
            new_source = DirectSum(point_map.source, param_space)

        self._bare = bare
        self._point_map = point_map
        self._param_names = param_names
        self._param_space = param_space
        super().__init__(new_source, point_map.target)

    def __call__(self, point_params, **kwargs):
        """
        Args:
            point_params: the point at which to evaluate the point map and the
                params to pass as keyword args. It must be in the form ``(point,
                params)`` (or ``(point, *params)`` if ``bare=True`` was passed
                in the constructor).

        Returns:
            The output of the base point map.
        """
        point = point_params[0]
        if self._bare:
            params = point_params[-self._param_len :]
        else:
            if len(point_params) != 2:
                raise ValueError(
                    "Expected a tuple of length 2 in the form (point, params),"
                    " but got a tuple of length %d. (Did you mean to use bare=True?)"
                    % len(point_params)
                )
            params = point_params[1]
            if not isinstance(self._param_space, DirectSum):
                params = [params]

        if len(params) != len(self._param_names):
            raise ValueError("")
        for name, val in zip(self._param_names, params):
            if name in kwargs:
                raise ValueError(
                    "%s specified as both a positional argument and keyword argument"
                    % name
                )
            kwargs[name] = val
        return self._point_map(point, **kwargs)


class Parametric(PointMap):
    """This class wraps a PointMap so that some parameters do not need to be
    specified when the point map is called.

    In mathematical notation, this wraps a function :math:`f(u, p)` so that it can be called as
    :math:`f(u)` or as :math:`f(u; p)` with some default value of p if p is not specified.

    Args:
        orig_map (PointMap): The point map to wrap.
        param_indices (int, tuple[int], or list[int]): The position of the
            parameters in the input space of the point map.
        param_point: The default values to use for the parameters if they are
            not specified in the :meth:`__call__` method.
        bare (bool): If you set this to true, then the number of remaining args
            to the point map (after removing the param_indices) must be 1, in
            which case the resulting Parametric map can be called as pmap(val)
            instead of pmap((val,)).
        bare_map (bool): If true, then the args to the point map will be
            unpacked before calling the map.
            Note: We should probably get rid of this since we're now requiring
            that PointMaps not accept bare arguments (I.e., all PointMap
            __call__ functions must accept a single arg, which is either a
            single value or a tuple of values if necessary).
    """

    def __init__(
        self,
        orig_map: PointMap,
        param_indices: Union[int, Iterable[int]],
        param_point: InputPoint,
        bare: Optional[bool] = False,
        bare_map: Optional[bool] = False,
    ):
        orig_source = orig_map.source
        if isinstance(param_indices, int):
            bare_param = True
            param_indices = (param_indices,)
        else:
            bare_param = False
        comp_indices = tuple(
            sorted(frozenset(range(len(orig_source))) - frozenset(param_indices))
        )
        if isinstance(orig_source, DirectSum):
            new_source = DirectSum(*tuple(orig_source[i] for i in comp_indices))
            param_space = DirectSum(*tuple(orig_source[i] for i in param_indices))
        elif isinstance(orig_source, Multiset):
            new_source = Multiset(
                orig_source.space, len(orig_source) - len(param_indices)
            )
            param_space = Multiset(orig_source.space, len(param_indices))
        if bare:
            assert len(new_source) == 1
            new_source = new_source[0]
        if bare_param:
            param_space = param_space[0]
        super(Parametric, self).__init__(new_source, orig_map.target)
        self._orig_map = orig_map
        self._comp_indices = comp_indices
        self._param_space = param_space
        self._param_indices = param_indices
        self._param_point = param_point
        self._bare_point = bare
        self._bare_param = bare_param
        self._bare_map = bare_map

    def __repr__(self):
        if self._bare_param:
            return f"Parametric({self._orig_map!r},arg[{self._param_indices[0]}]={self._param_point})"
        else:
            return f"Parametric({self._orig_map!r},arg[{self._param_indices}]={self._param_point})"

    def __call__(
        self, point: InputPoint, params: Optional[ParameterPoint] = None, **kwargs
    ) -> OutputPoint:
        """
        Args:
            point: point at which to evaluate the point map
            params: parameter values to use for the point map. If None, then the
                default param_point will be used.
            **kwargs: keyword arguments are passed through to the base point map.

        Returns:
            The output of the base point map.
        """
        params = self._param_point if not params else params
        full = [None] * len(self._orig_map.source)
        if self._bare_point:
            point = (point,)
        if self._bare_param:
            params = (params,)
        for i, j in enumerate(self._param_indices):
            full[j] = params[i]
        for i, j in enumerate(self._comp_indices):
            full[j] = point[i]
        full = tuple(full)
        if self._bare_map:
            return self._orig_map(*full, **kwargs)
        else:
            return self._orig_map(full, **kwargs)

    def set_param_point(self, param_point: ParameterPoint) -> None:
        """Sets the default value to use for the parameters."""
        self._param_point = param_point


class CompositePointMap(PointMap):
    """This class is a point map that links a group of point maps.

    Given a list of point maps, the source space is the source space of the
    first point map, and the target space is the target space of the last point
    map.

    The __call__ function gives the inputs to the first map, and then
    successively feeds the outputs of one map to the next map.

    """

    def __init__(self, *point_maps):
        # This allows CompositePointMap(map_list) or CompositePointMap(*map_list).
        if len(point_maps) == 1 and isinstance(point_maps[0], (list, tuple)):
            point_maps = point_maps[0]
        self._point_maps = point_maps

        source = self._point_maps[0].source
        target = self._point_maps[-1].target
        super().__init__(source, target)

    def __repr__(self):
        return f"CompositePointMap{tuple(self._point_maps)}"

    def __call__(self, args: InputPoint) -> OutputPoint:
        for point_map in self._point_maps:
            args = point_map(args)
        return args

    def point_maps(self) -> List[PointMap]:
        """Returns a list of the point maps used to create this composite map."""
        return self._point_maps


class ParallelPointMap(PointMap):
    """This class is a point map that runs a group of point maps independently.

    Given a list of point maps, the source space is a DirectSum of the source
    space from each point map, and similarly for the target space.

    The __call__ function gives each argument to the corresponding point map,
    and concatenates the outputs into a single list.

    """

    def __init__(self, *point_maps):
        if len(point_maps) == 1 and isinstance(point_maps, (tuple, list)):
            point_maps = tuple(point_maps[0])
        self._point_maps = point_maps
        source = DirectSum(*tuple(point_map.source for point_map in self._point_maps))
        target = DirectSum(*tuple(point_map.target for point_map in self._point_maps))

        super().__init__(source, target)

    def __call__(self, arg):
        return tuple(p(a) for p, a in zip(self._point_maps, arg))


class IdentityPointMap(PointMap):
    """This class is a point map that simply returns its input."""

    def __call__(self, point: InputPoint, **kwargs) -> InputPoint:
        return point


class FunnelOut(PointMap):
    """Makes n copies of its input"""

    def __init__(self, source_space, n):
        self._n = n
        target_space = DirectSum((source_space,) * n)
        super().__init__(source_space, target_space)

    def __call__(self, arg: InputPoint) -> OutputPoint:
        return (arg,) * self._n



from crikit.fe import *
from crikit.fe_adjoint import *
from pyadjoint.tape import annotate_tape, get_working_tape, stop_annotating
from pyadjoint.overloaded_type import create_overloaded_object
import jax
from jax import numpy as jnp
from jax import random
import numpy as np
from pyadjoint_utils import Block
from pyadjoint_utils.jax_adjoint import array, ndarray, overload_jax
from pyadjoint.tape import annotate_tape, stop_annotating, get_working_tape
from crikit.cr.jax_utils import (
    JAXArrays,
    UFLExprSpace_To_JAX,
    JAX_UFLFunctionSpace_Covering,
)
from crikit.cr.ufl import UFLExprSpace
from crikit.cr.types import PointMap
from crikit.covering import get_composite_cr
from typing import Optional, List


def u_observer(u):
    return u.copy(deepcopy=True)


class SurfaceObserver:
    class SurfaceObservation:
        def __init__(self, u, de):
            self.u = u.copy(deepcopy=True)
            self.de = de

    def __init__(self, de):
        self.de = de

    def __call__(self, u):
        return self.SurfaceObservation(u, self.de)


class SubdomainObserver:
    def __init__(self, mesh: Mesh, subdomain: SubDomain):
        """An Observer that zeros all values of a :class:`Function` outside of a certain :class:`SubDomain`.

        :param mesh: The mesh
        :type mesh: (~fenics.Mesh)
        :param subdomain: The subdomain
        :type subdomain: SubDomain
        :return: The SubdomainObserver
        :rtype: SubdomainObserver
        """
        self._cell_func = MeshFunction("size_t", mesh, mesh.topology().dim(), 0)
        self._subdomain = subdomain
        self._subdomain.mark(self._cell_func, 1)

    def __call__(self, u: Function) -> Function:
        """Compute the observation by zeroing out values off of the SubDomain.

        :param u: The :class:`Function` to observe
        :type u: Function
        :return: The observed `u`
        :rtype: Function
        """
        annotate = annotate_tape()
        with stop_annotating():
            obs = self._zero_complement(u.copy(deepcopy=True), 1)
        if annotate_tape:
            obs = create_overloaded_object(obs)
            block = SubdomainObserverBlock(create_overloaded_object(u), obs, self)
            tape = get_working_tape()
            tape.add_block(block)

        return obs

    def _zero_complement(self, u, subdomain_id):
        u_vec_cpy = self._zero_complement_vec(u, u.vector(), subdomain_id)
        u.vector().set_local(u_vec_cpy.get_local())
        return u

    def _zero_complement_vec(self, u_func, u_vec, subdomain_id):
        # create a Function like u that's 1 everywhere on the subdomain, 0 off of it
        # and use that to determine which indices of u to zero out
        ucb = (
            Constant((1,) * u_func.ufl_shape[0])
            if len(u_func.ufl_shape) > 0
            else Constant(1)
        )
        dbc = DirichletBC(u_func.function_space(), ucb, self._subdomain)
        v = Function(u_func.function_space())
        dbc.apply(v.vector())
        idx = np.where(v.vector().vec().array != 1)
        uc = u_vec.copy()
        u_locals = uc.get_local()
        u_locals[idx] = 0.0
        uc.set_local(u_locals)
        return uc


class SubdomainObserverBlock(Block):
    def __init__(self, var, output, observer):
        super().__init__()
        self.observer = observer
        self.add_dependency(var)
        self.add_output(output.create_block_variable())

    def recompute_component(self, inputs, block_variable, idx, prepared):
        return self.observer._zero_complement(inputs[0].copy(deepcopy=True), 1)

    def evaluate_adj_component(self, inputs, adj_inputs, block_variable, idx, prepared):
        return self.observer._zero_complement_vec(inputs[0], adj_inputs[0], 1)

    def evaluate_tlm_component(self, inputs, tlm_inputs, block_variable, idx, prepared):
        return self.observer._zero_complement(tlm_inputs[0], 1)

    def evaluate_hessian_component(
        self, inputs, hessian_inputs, adj_inputs, block_variable, prepared
    ):
        return self.evaluate_adj_component(
            inputs, hessian_inputs, block_variable, idx, prepared
        )


class AdditiveRandomFunction:
    """A class representing a random function on a mesh that adds a sample
    from its distribution to the inputs (i.e. an additive noise model). In
    other words, if ``X`` is your UFL input and ``Y`` is the random variable
    this class represents, then this class returns ``X + Y``.
    """

    def __init__(
        self,
        V: FunctionSpace,
        distribution: Optional[str] = "normal",
        seed: Optional[int] = 0,
        **kwargs,
    ):
        """
        :param V: The ``FunctionSpace`` in which to generate noise
        :type V: FunctionSpace
        :param distribution: A string describing the distribution of this function. Use the static method :meth:`AdditiveRandomFunction.available_distributions` to see available distributions, defaults to 'normal'
        :type distribution: str, optional
        :param seed: the seed for ``jax.random``, defaults to 0
        :type seed: int, optional
        :param kwargs: Keyword arguments (e.g. parameter values) to be passed
            on to the distribution. For example, if ``distribution`` is 'gamma',
            you should pass ``a = value_of_a``, since ``jax.random.gamma`` takes
            a parameter named ``a``.
        :type kwargs: dict, optional

        """
        self._V = V
        self._key = random.PRNGKey(seed)
        self._kwargs = kwargs
        dmap = self._get_distribution_map()
        if distribution not in dmap:
            raise ValueError(
                f"Distribution {distribution} is not in the list of available distributions! The list of available distributions is \n{list(dmap.keys())}."
            )

        self._distribution = dmap[distribution.lower()]
        self._dname = distribution.lower()

    def __str__(self):
        return (
            f"AdditiveRandomFunction(distribution={self._dname}, kwargs={self._kwargs})"
        )

    def __call__(self, ufl_input: Function) -> Function:
        """Generate and add the noise.

        :param ufl_input: The input
        :type ufl_input: Function
        :return: The input plus some random noise
        :rtype: Function
        """
        annotate = annotate_tape()
        with stop_annotating():
            noise_f = Function(self._V)
            self._kwargs["shape"] = noise_f.vector().vec().array.shape
            self._key, subkey = random.split(self._key)
            noise = self._distribution(subkey, **self._kwargs)
            noisy_ufl = ufl_input.copy(deepcopy=True)
            noisy_ufl.vector()[:] += noise

        if annotate:
            tape = get_working_tape()
            tape.add_block(AdditiveRandomBlock(ufl_input, noisy_ufl, noise))

        return noisy_ufl

    @staticmethod
    def _normal(key, **kwargs):
        std = kwargs.pop("std", 1.0)
        mu = kwargs.pop("mu", 0.0)
        return std * random.normal(key, **kwargs) + mu

    @staticmethod
    def _get_distribution_list() -> List[str]:
        """Returns a list of available distributions
        :return: list of strings that can be passed to the constructor as distributions
        :rtype: list
        """
        return list(AdditiveRandomFunction._get_distribution_map().keys())

    @staticmethod
    def get_available_params() -> dict:
        """
        Returns a dictionary mapping distribution names to the names of
        parameters for that distribution (pass as kwargs to the constructor for
        this class). For example, if you're using a ``pareto`` distribution,
        you'll also want to pass ``b=your_b_value`` to the constructor of this class.

        Currently, that dictionary is
        ::

            {'bernoulli' : 'p',
                'beta' : 'a, b',
                'dirichlet' : 'alpha',
                'double_sided_maxwell' : 'loc, scale',
                'gamma' : 'a',
                'multivariate_normal' : 'mean, cov',
                'normal' : 'mu, std',
                'pareto' : 'b',
                'poisson' : 'lam',
                't' : 'df',
                'truncated_normal' : 'lower, upper',
                'uniform' : 'minval, maxval',
                'weibull' : 'concentration'
             }

        """
        return {
            "bernoulli": "p",
            "beta": "a, b",
            "dirichlet": "alpha",
            "double_sided_maxwell": "loc, scale",
            "gamma": "a",
            "multivariate_normal": "mean, cov",
            "normal": "mu, std",
            "pareto": "b",
            "poisson": "lam",
            "t": "df",
            "truncated_normal": "lower, upper",
            "uniform": "minval, maxval",
            "weibull": "concentration",
        }

    @staticmethod
    def _get_distribution_map():
        return {
            "bernoulli": random.bernoulli,
            "cauchy": random.cauchy,
            "dirichlet": random.dirichlet,
            "double_sided_maxwell": random.double_sided_maxwell,
            "exponential": random.exponential,
            "gamma": random.gamma,
            "gumbel": random.gumbel,
            "laplace": random.laplace,
            "logistic": random.logistic,
            "maxwell": random.maxwell,
            "multivariate_normal": random.multivariate_normal,
            "normal": AdditiveRandomFunction._normal,
            "pareto": random.pareto,
            "poisson": random.poisson,
            "rademacher": random.rademacher,
            "t": random.t,
            "truncated_normal": random.truncated_normal,
            "uniform": random.uniform,
            "weibull": random.weibull_min,
        }

    @staticmethod
    def available_distributions() -> List[str]:
        return AdditiveRandomFunction._get_distribution_list()


class AdditiveRandomBlock(Block):
    def __init__(self, var, output, noise):
        super().__init__()
        self.noise = noise
        self.add_dependency(var)
        self.add_output(output.create_block_variable())

    def recompute_component(self, inputs, block_variable, idx, prepared):
        var = inputs[0].copy(deepcopy=True)
        var.vector()[:] += self.noise
        return var

    def evaluate_adj_component(
        self, inputs, adj_inputs, block_variable, idx, prepared=None
    ):
        return adj_inputs[0]

    def evaluate_tlm_component(
        self, inputs, tlm_inputs, block_variable, idx, prepared=None
    ):
        return tlm_inputs[0]

    def evaluate_tlm_matrix_component(
        self, inputs, tlm_inputs, block_variable, idx, prepared=None
    ):
        return inputs[0]

    def evaluate_hessian_component(
        self,
        inputs,
        hessian_inputs,
        adj_inputs,
        block_variable,
        idx,
        relevant_dependencies,
        prepared=None,
    ):
        return hessian_inputs[0]



"""This submodule defines a default FE backend (FEniCS or Firedrake)"""

from ufl import indices, as_tensor

try:
    from firedrake import *

    crikit_fe_backend = "firedrake"
except ModuleNotFoundError:
    try:
        from fenics import *

        crikit_fe_backend = "fenics"
        from ffc.quadrature.deprecation import (
            QuadratureRepresentationDeprecationWarning,
        )
        import warnings

        warnings.simplefilter("ignore", QuadratureRepresentationDeprecationWarning)
        del warnings, QuadratureRepresentationDeprecationWarning
        dx = dx(metadata={"representation": "quadrature"})
    except ModuleNotFoundError:
        raise ImportError("Could not import a FE backend")


def mesh_from_ufl_domain(domain):
    if crikit_fe_backend == "firedrake":
        return domain
    else:
        return domain.ufl_cargo()


def contraction(a, a_axes, b, b_axes):
    "UFL operator: Take the contraction of a and b over given axes."
    ai, bi = a_axes, b_axes
    if len(ai) != len(bi):
        error("Contraction must be over the same number of axes.")
    ash = a.ufl_shape
    bsh = b.ufl_shape
    aii = list(indices(len(a.ufl_shape)))  ######### Convert tuple to list
    bii = list(indices(len(b.ufl_shape)))  ######### Convert tuple to list
    cii = indices(len(ai))
    shape = [None] * len(ai)
    for i, j in enumerate(ai):
        aii[j] = cii[i]
        shape[i] = ash[j]
    for i, j in enumerate(bi):
        bii[j] = cii[i]
        if shape[i] != bsh[j]:
            error("Shape mismatch in contraction.")
    aii = tuple(aii)  ######### Convert list back to tuple
    bii = tuple(bii)  ######### Convert list back to tuple
    s = a[aii] * b[bii]
    cii = set(cii)
    ii = tuple(i for i in (aii + bii) if i not in cii)
    return as_tensor(s, ii)



__author__ = "CRIKit Team"
from ._version import __version__

import os

_jax_enable_x64 = os.getenv("JAX_ENABLE_X64")
if _jax_enable_x64 is None or "true" in _jax_enable_x64.lower():
    from jax.config import config as _jax_config

    if not _jax_config.read("jax_enable_x64"):
        _jax_config.update("jax_enable_x64", True)
    del _jax_config

del os

from . import fe
from . import fe_adjoint
from .fe import *
from .fe_adjoint import *
from . import cr
from .cr import (
    CR,
    BlockCR,
    cr_function_shape,
    JAXArrays,
    JAX_To_UFLFunctionSpace,
    UFLExprSpace_To_JAX,
    JAX_UFLFunctionSpace_Covering,
    ReducedFunctionJAX,
    UFLFunctionSpace,
    UFLExprSpace,
    CR_UFL_Expr,
    create_ufl_standins,
    create_ufl_standin_arguments,
    Ndarrays,
    form_get_expr_space,
    assemble_with_cr,
    get_ufl_composite_cr,
    get_cr_form_degree,
    AssembleWithCRBlock,
    Space,
    PointMap,
    DirectSum,
    Callable,
    AugmentPointMap,
    Parametric,
    CompositePointMap,
    ParallelPointMap,
    IdentityPointMap,
)
from . import invariants
from .invariants import (
    symm,
    antisymm,
    eps_ij,
    eps_ijk,
    get_invariant_functions,
    get_invariant_descriptions,
    LeviCivitaType,
    register_invariant_functions,
    InvariantInfo,
    TensorType,
    levi_civita,
)
from . import covering
from .covering import (
    Covering,
    get_composite_cr,
    get_map,
    reset_default_covering_params,
    register_covering,
    get_default_covering_params,
    set_default_covering_params,
    Numpy_UFLFunctionSpace_Covering,
    Numpy_To_UFLFunctionSpace,
    UFLExprSpace_To_Numpy,
    UFLFunctionSpace_UFLExpr_Covering,
    To_UFLFunctionSpace,
    UFLFunctionSpace_UFLFunctionSpace_Covering,
)

from .projection import project, ProjectBlock
from .loss import integral_loss, vector_loss, SlicedWassersteinDistance
from .observer import SurfaceObserver, SubdomainObserver, AdditiveRandomFunction
from . import logging
from .logging import set_log_level



from pyadjoint.overloaded_function import overload_function

# from pyadjoint.tape import get_working_tape, set_working_tape, stop_annotating, annotate_tape
from pyadjoint.overloaded_type import create_overloaded_object
from pyadjoint_utils import Block
import numpy as np
import jax
from jax import numpy as jnp
from fenics_adjoint import backend
from pyadjoint_utils.numpy_adjoint import *


def backend_function_get_local(f):
    return f.vector().get_local().reshape((-1, *f.ufl_shape))


def backend_function_set_local(f, arr):
    a = np.asarray(arr.flatten())
    f.vector().set_local(a)
    if len(f.vector()) != len(a):
        raise ValueError(
            f"Incorrect array size: {len(f.vector())} != {len(np.asarray(arr.flatten()))}"
        )
    return f


class FunctionGetLocalBlock(Block):
    def __init__(self, f):
        super().__init__()
        self.V = f.function_space()
        self.add_dependency(f)

    def __str__(self):
        return "FunctionGetLocalBlock"

    def evaluate_adj_component(
        self, inputs, adj_inputs, block_variable, idx, prepared=None
    ):
        f = block_variable.saved_output
        arr = adj_inputs[0]
        adj_value = backend_function_set_local(backend.Function(self.V), arr)
        return adj_value.vector()

    def evaluate_tlm_component(
        self, inputs, tlm_inputs, block_variable, idx, prepared=None
    ):
        f = tlm_inputs[0]
        return backend_function_get_local(f)

    def evaluate_tlm_matrix_component(
        self, inputs, tlm_inputs, block_variable, idx, prepared=None
    ):
        return tlm_inputs[0]

    def recompute_component(self, inputs, block_variable, idx, prepared):
        f = inputs[0]
        return backend_function_get_local(f)


class FunctionSetLocalBlock(Block):
    def __init__(self, f, arr):
        super().__init__()
        self.V = f.function_space()
        self.add_dependency(arr)

    def __str__(self):
        return "FunctionSetLocalBlock"

    def evaluate_adj_component(
        self, inputs, adj_inputs, block_variable, idx, prepared=None
    ):
        arr = block_variable.saved_output
        fvec = adj_inputs[0]
        return fvec.get_local().reshape(arr.shape)

    def evaluate_tlm_component(
        self, inputs, tlm_inputs, block_variable, idx, prepared=None
    ):
        arr = tlm_inputs[0]
        return function_set_local(backend.Function(self.V), arr)

    def evaluate_tlm_matrix_component(
        self, inputs, tlm_inputs, block_variable, idx, prepared=None
    ):
        return tlm_inputs[0]

    def recompute_component(self, inputs, block_variable, idx, prepared):
        arr = inputs[0]
        return backend_function_set_local(backend.Function(self.V), arr)


function_get_local = overload_function(
    backend_function_get_local, FunctionGetLocalBlock
)
function_set_local = overload_function(
    backend_function_set_local, FunctionSetLocalBlock
)



from crikit.cr.types import PointMap
from pyadjoint.enlisting import Enlist
from pyadjoint.overloaded_type import create_overloaded_object, OverloadedType
from pyadjoint_utils.numpy_adjoint.autograd import overloaded_autograd
from scipy.stats import special_ortho_group
from ufl.operators import as_matrix, dot, transpose
import numpy as np
from typing import Iterable, List, Optional


class SpecialOrthoGroup:
    """This class represents the special orthogonal group, which is the group of
    arbitrary rotations without reflections.

    Args:
        dim (int): The dimension of the group, which is usually 2 or 3.
    """

    def __init__(self, dim: int):
        self.dim = dim

    def sample(self) -> OverloadedType:
        """This method returns a sample of the group."""
        return create_overloaded_object(special_ortho_group.rvs(self.dim))

    def apply(self, sample: OverloadedType, tensors: List[np.ndarray], offset=1):
        """This method applies a sample of the group to a group of tensors.

        Each tensor is assumed to be in the form of a NumPy array. The first
        axis of each array is ignored, and the transformation is applied to the
        remaining axes.

        Args:
            sample: a sample from the group.
            tensors: a tensor or list of tensors to apply the transformation to.

        Returns:
            type: returns the transformed tensor(s).

        """
        tensors = Enlist(tensors)
        transformed = [
            apply_coordinate_change(sample, T, offset=offset) for T in tensors
        ]
        return tensors.delist(transformed)


@overloaded_autograd(pointwise=(False, True))
def apply_coordinate_change(Q, T, offset=1):
    """Applies the coordinate transformation matrix Q to the tensor T."""
    shape = T.shape[offset:]
    rank = len(shape)
    covariants = range(rank)
    contravariants = ()

    einsum_args = get_einsum_args(
        rank, contravariants, covariants, Q, None, offset=offset
    )
    return np.einsum(T, *einsum_args)


def get_einsum_args(
    rank: int,
    contravariants: List[int],
    covariants: List[int],
    basis: np.ndarray,
    inv_basis: np.ndarray,
    offset: Optional[int] = 0,
):
    """Generates arguments for a NumPy's einsum function so that calling that
    function with these arguments and a tensor will perform a tensor coordinate
    transformation.

    The tensor is of the given rank, with the given covariant axes and
    contravariants axes. Each specified axis should be in the range [0, rank-1].

    If the coordinate transformation is being performed on a set of tensors
    stored together in the same NumPy array, the first axis (or first few axes)
    of the array is not part of the tensor definition. The offset argument
    specifies how many of the first axes to ignore. For example, if N tensors
    each with shape S are stored in an array of shape (N, S), the offset should
    be set to 1.

    Args:
        rank (int): the rank of the tensor to be transformed.
        contravariants (list): the axes that transform contravariantly.
        covariants (list): the axes that transform covariantly.
        basis: the coordinate basis to use for the transform.
        inv_basis: the inverse of the coordinate basis, used for the contravariant axes.
        offset(int): the first axes will be unchanged.

    Returns:
        tuple: returns a tuple of args to be used with NumPy's einsum function.

    """
    # Offset gives the number of axes that will be ignored.
    num_axes = rank + offset

    # Set the input axes as the full number of axes.
    axes = tuple(range(num_axes))
    args = [axes]

    # Apply the appropriate transformation to each axis.
    for ax in contravariants:
        args.append(inv_basis)
        args.append([ax + offset, ax + num_axes])
    for ax in covariants:
        args.append(basis)
        args.append([ax + offset, ax + num_axes])

    # The output axes are the ignored axes plus the transformed axes.
    new_axes = tuple(range(offset)) + tuple(range(num_axes, num_axes + rank))
    args.append(new_axes)

    return args


class GroupAction(PointMap):
    """This point map applies the action of a group to the given inputs.

    Args:
        group: The group that will be used to perform the action.
        space (Space): The space that the group transformations will be applied to.
        sample: a sample of the group that will be used if no sample is
            specified in the :meth:`__call__` method.
    """

    def __init__(self, group, space, sample=None):
        self.group = group
        self.sample = None
        super().__init__(space, space)

    def __call__(self, inputs, sample=None):
        """Applies the group transformation to the given inputs using the given
        group sample. If no sample is given, the default sample is used."""
        if sample is None:
            sample = self.sample
        return self.group.apply(sample, inputs)



from .types import PointMap
from .space_builders import DirectSum, enlist
from .numpy import Ndarrays
from .overloaded import overloaded_point_map
from autograd import vector_jacobian_product, jacobian, elementwise_grad
from autograd.numpy.numpy_boxes import ArrayBox
from autograd.numpy.numpy_vspaces import ArrayVSpace
from pyadjoint.enlisting import Enlist
from pyadjoint import AdjFloat
from pyadjoint_utils.numpy_adjoint import ndarray
from pyadjoint_utils.numpy_adjoint.autograd import overload_autograd
from functools import wraps
from itertools import product
import numpy as np
import autograd.numpy as anp

ArrayBox.register(ndarray)
ArrayBox.register(AdjFloat)
ArrayVSpace.register(ndarray)
ArrayVSpace.register(AdjFloat)

from autograd.extend import defvjp


def vjpmaker_trace(ans, x, offset=0, axis1=0, axis2=1):
    axes = tuple(range(x.ndim))
    eye = anp.eye(x.shape[axis1], x.shape[axis2], k=offset, dtype=bool)
    eye = anp.where(eye, anp.ones_like(x), anp.zeros_like(x))

    def vjp(g):
        g = anp.expand_dims(g, axis=axis1)
        g = anp.expand_dims(g, axis=axis2)
        return eye * g

    return vjp


defvjp(anp.trace, vjpmaker_trace)


class AutogradPointMap(PointMap):
    def __init__(self, source, target, func, bare=False, pointwise=True):
        self._orig_func = func
        self._bare = bare

        # Autograd requires once one simple type (ndarray/stdnumeric) per *args index: fit ag_func to that
        if self._bare:
            self._ag_func = self._orig_func
        else:

            @wraps(self._orig_func)
            def _expanded_func(*args, **kwargs):
                return self._orig_func(args, **kwargs)

            self._ag_func = _expanded_func

        self._overloaded_func = overload_autograd(self._ag_func, pointwise)
        self.pointwise = self._overloaded_func.pointwise
        super().__init__(source, target)

    def __repr__(self):
        return f"AutogradPointMap({self._orig_func})"

    def __call__(self, arg, **kwargs):
        if isinstance(self.source, DirectSum):
            return self._overloaded_func(*arg, **kwargs)
        else:
            return self._overloaded_func(arg, **kwargs)


def point_map(source_tuple, target_tuple, **kwargs):
    """This decorator turns the decorated function into an AutogradPointMap.

    The given tuples are used to create :class:`~crikit.cr.numpy.Ndarrays` spaces to set as the
    source and target spaces.
    """
    dtype = kwargs.get("dtype", None)
    if source_tuple is None:
        source = None
    elif any(isinstance(s, tuple) for s in source_tuple):
        source = DirectSum(*tuple(Ndarrays(s, dtype) for s in source_tuple))
    else:
        source = Ndarrays(source_tuple, dtype)
    if any(isinstance(t, tuple) for t in target_tuple):
        target = DirectSum(*tuple(Ndarrays(t, dtype) for t in target_tuple))
    else:
        target = Ndarrays(target_tuple, dtype)

    def point_map_decorator(func):
        ag = AutogradPointMap(source, target, func, **kwargs)
        return ag

    return point_map_decorator



import os

_jax_enable_x64 = os.getenv("JAX_ENABLE_X64")
if _jax_enable_x64 is None or not ("true" in _jax_enable_x64.lower()):
    import jax

    jax.config.update("jax_enable_x64", True)
    del jax

del _jax_enable_x64
del os

from .array import (
    ndarray,
    array,
    asarray,
    to_jax,
    to_adjfloat,
    set_default_dtype,
    get_default_dtype,
)
from .jax import overload_jax



from fenics_adjoint import backend, compat
from pyadjoint.enlisting import Enlist


def homogenize_bcs(bcs):
    bcs = Enlist(bcs)
    h_bcs = []
    for bc in bcs:
        if isinstance(bc, backend.DirichletBC):
            bc = compat.create_bc(bc, homogenize=True)
        h_bcs.append(bc)
    return bcs.delist(h_bcs)
